@@ -16,8 +16,7 @@ struct spotify_backend
 {
   int (*init)(void);
   void (*deinit)(void);
-  int (*login)(const char *username, const char *password, const char **errmsg);
-  int (*login_token)(const char *username, const char *token, const char **errmsg);
+  int (*login)(const char *username, const char *token, const char **errmsg);
   void (*logout)(void);
   int (*relogin)(void);
   void (*uri_register)(const char *uri);
@@ -31,10 +30,7 @@ void
 spotify_deinit(void);
 
 int
-spotify_login(const char *username, const char *password, const char **errmsg);
-
-int
-spotify_login_token(const char *username, const char *token, const char **errmsg);
+spotify_login(const char *username, const char *token, const char **errmsg);
 
 void
 spotify_logout(void);
