@@ -20,6 +20,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdbool.h>
 #include <unistd.h>
 #include <stdint.h>
 #include <fcntl.h>
@@ -109,6 +110,7 @@ static int
 postlogin(struct global_ctx *ctx)
 {
   struct sp_credentials credentials;
+  bool use_legacy_mode;
   char *db_stored_cred;
   char *ptr;
   int i;
@@ -135,7 +137,13 @@ postlogin(struct global_ctx *ctx)
 
   librespotc_bitrate_set(ctx->session, ctx->bitrate_preferred);
 
-  DPRINTF(E_LOG, L_SPOTIFY, "Logged into Spotify succesfully with username %s\n", credentials.username);
+  // For now, use old tcp based protocol as default unless configured not to.
+  // Note that setup() will switch the old protocol off on error.
+  use_legacy_mode = !cfg_getbool(cfg_getsec(cfg, "spotify"), "disable_legacy_mode");
+  if (use_legacy_mode)
+    librespotc_legacy_set(ctx->session, true);
+
+  DPRINTF(E_LOG, L_SPOTIFY, "Logged into Spotify succesfully with username %s (%s mode)\n", credentials.username, use_legacy_mode ? "tcp" : "http");
 
   listener_notify(LISTENER_SPOTIFY);
 
@@ -217,40 +225,6 @@ progress_cb(int fd, void *cb_arg, size_t received, size_t len)
   DPRINTF(E_SPAM, L_SPOTIFY, "Progress %zu/%zu\n", received, len);
 }
 
-static int
-https_get_cb(char **out, const char *url)
-{
-  struct http_client_ctx ctx = { 0 };
-  char *body;
-  size_t len;
-  int ret;
-
-  ctx.url = url;
-  ctx.input_body = evbuffer_new();
-
-  ret = http_client_request(&ctx, NULL);
-  if (ret < 0 || ctx.response_code != HTTP_OK)
-    {
-      DPRINTF(E_LOG, L_SPOTIFY, "Failed to AP list from '%s' (return %d, error code %d)\n", ctx.url, ret, ctx.response_code);
-      goto error;
-    }
-
-  len = evbuffer_get_length(ctx.input_body);
-  body = malloc(len + 1);
-
-  evbuffer_remove(ctx.input_body, body, len);
-  body[len] = '\0'; // For safety
-
-  *out = body;
-
-  evbuffer_free(ctx.input_body);
-  return 0;
-
- error:
-  evbuffer_free(ctx.input_body);
-  return -1;
-}
-
 static int
 tcp_connect(const char *address, unsigned short port)
 {
@@ -289,7 +263,6 @@ hexdump_cb(const char *msg, uint8_t *data, size_t data_len)
 /* ------------------------ librespot-c initialization ---------------------- */
 
 struct sp_callbacks callbacks = {
-  .https_get      = https_get_cb,
   .tcp_connect    = tcp_connect,
   .tcp_disconnect = tcp_disconnect,
 
@@ -391,21 +364,18 @@ download_seek(void *arg, int64_t offset, enum transcode_seek_type type)
 static int
 download_xcode_setup(struct download_ctx *download)
 {
-  struct transcode_ctx *xcode;
+  struct transcode_decode_setup_args decode_args = { .profile = XCODE_OGG, .len_ms = download->len_ms };
+  struct transcode_encode_setup_args encode_args = { .profile = XCODE_PCM16, };
   struct transcode_evbuf_io xcode_evbuf_io = { 0 };
-
-  CHECK_NULL(L_SPOTIFY, xcode = malloc(sizeof(struct transcode_ctx)));
+  struct transcode_ctx *xcode;
 
   xcode_evbuf_io.evbuf = download->read_buf;
   xcode_evbuf_io.seekfn = download_seek;
   xcode_evbuf_io.seekfn_arg = download;
+  decode_args.evbuf_io = &xcode_evbuf_io;
 
-  xcode->decode_ctx = transcode_decode_setup(XCODE_OGG, NULL, DATA_KIND_SPOTIFY, NULL, &xcode_evbuf_io, download->len_ms);
-  if (!xcode->decode_ctx)
-    goto error;
-
-  xcode->encode_ctx = transcode_encode_setup(XCODE_PCM16, NULL, xcode->decode_ctx, NULL, 0, 0);
-  if (!xcode->encode_ctx)
+  xcode = transcode_setup(decode_args, encode_args);
+  if (!xcode)
     goto error;
 
   download->xcode = xcode;
@@ -524,6 +494,10 @@ setup(struct input_source *source)
   return 0;
 
  error:
+  // This should be removed when we don't default to legacy mode any more
+  DPRINTF(E_INFO, L_SPOTIFY, "Switching off Spotify legacy mode\n");
+  librespotc_legacy_set(ctx->session, false);
+
   pthread_mutex_unlock(&spotify_ctx_lock);
   stop(source);
 
@@ -644,40 +618,7 @@ struct input_definition input_spotify =
 /*             Called from other threads than the input thread                */
 
 static int
-login(const char *username, const char *password, const char **errmsg)
-{
-  struct global_ctx *ctx = &spotify_ctx;
-  int ret;
-
-  pthread_mutex_lock(&spotify_ctx_lock);
-
-  ctx->session = librespotc_login_password(username, password);
-  if (!ctx->session)
-    goto error;
-
-  ret = postlogin(ctx);
-  if (ret < 0)
-    goto error;
-
-  pthread_mutex_unlock(&spotify_ctx_lock);
-
-  return 0;
-
- error:
-  if (ctx->session)
-    librespotc_logout(ctx->session);
-  ctx->session = NULL;
-
-  if (errmsg)
-    *errmsg = librespotc_last_errmsg();
-
-  pthread_mutex_unlock(&spotify_ctx_lock);
-
-  return -1;
-}
-
-static int
-login_token(const char *username, const char *token, const char **errmsg)
+login(const char *username, const char *token, const char **errmsg)
 {
   struct global_ctx *ctx = &spotify_ctx;
   int ret;
@@ -785,7 +726,6 @@ status_get(struct spotify_status *status)
 struct spotify_backend spotify_librespotc =
 {
   .login = login,
-  .login_token = login_token,
   .logout = logout,
   .relogin = relogin,
   .status_get = status_get,
