@@ -29,9 +29,6 @@
 #include <stdbool.h>
 #include <stdlib.h>
 #include <string.h>
-#include <unictype.h>
-#include <uninorm.h>
-#include <unistr.h>
 
 #include <event2/event.h>
 
@@ -70,6 +67,14 @@ struct queue_item_add_param
   int *new_item_id;
 };
 
+struct item_param
+{
+  const char *path;
+  uint32_t id;
+  enum library_attrib attrib;
+  uint32_t value;
+};
+
 static struct commands_base *cmdbase;
 static pthread_t tid_library;
 
@@ -122,10 +127,12 @@ static struct library_callback_register library_cb_register[LIBRARY_MAX_CALLBACK
 int
 library_media_save(struct media_file_info *mfi)
 {
-  if (!mfi->path || !mfi->fname)
+  int ret;
+
+  if (!mfi->path || !mfi->fname || !mfi->scan_kind)
     {
-      DPRINTF(E_LOG, L_LIB, "Ignoring media file with missing values (path='%s', fname='%s', data_kind='%d')\n",
-	      mfi->path, mfi->fname, mfi->data_kind);
+      DPRINTF(E_LOG, L_LIB, "Ignoring media file with missing values (path='%s', fname='%s', scan_kind='%d', data_kind='%d')\n",
+	      mfi->path, mfi->fname, mfi->scan_kind, mfi->data_kind);
       return -1;
     }
 
@@ -137,26 +144,25 @@ library_media_save(struct media_file_info *mfi)
     }
 
   if (mfi->id == 0)
-    return db_file_add(mfi);
+    ret = db_file_add(mfi);
   else
-    return db_file_update(mfi);
+    ret = db_file_update(mfi);
+
+  return ret;
 }
 
 int
 library_playlist_save(struct playlist_info *pli)
 {
-  if (!pli->path)
+  if (!pli->path || !pli->scan_kind)
     {
-      DPRINTF(E_LOG, L_LIB, "Ignoring playlist file with missing path\n");
+      DPRINTF(E_LOG, L_LIB, "Ignoring playlist with missing values (path='%s', scan_kind='%d')\n",
+	      pli->path, pli->scan_kind);
       return -1;
     }
 
-  if (!pli->directory_id || !pli->virtual_path)
-    {
-      // Missing informations for virtual_path and directory_id (may) lead to misplaced appearance in mpd clients
-      DPRINTF(E_WARN, L_LIB, "Playlist with missing values (path='%s', directory='%d', virtual_path='%s')\n",
-	      pli->path, pli->directory_id, pli->virtual_path);
-    }
+  // Missing virtual_path and directory_id (may) lead to misplaced appearance in
+  // mpd clients, but for e.g. spotify:playlistfolder they will not be set
 
   if (pli->id == 0)
     return db_pl_add(pli);
@@ -164,6 +170,39 @@ library_playlist_save(struct playlist_info *pli)
     return db_pl_update(pli);
 }
 
+int
+library_directory_save(char *virtual_path, char *path, int disabled, int parent_id, enum scan_kind scan_kind)
+{
+  struct directory_info di = { 0 };
+  int id;
+  int ret;
+
+  id = db_directory_id_byvirtualpath(virtual_path);
+
+  di.id = id;
+  di.parent_id = parent_id;
+  di.virtual_path = safe_strdup(virtual_path);
+  di.path = safe_strdup(path);
+  di.disabled = disabled;
+  di.db_timestamp = (uint64_t)time(NULL);
+  di.scan_kind = scan_kind;
+
+  if (di.id == 0)
+    ret = db_directory_add(&di, &id);
+  else
+    ret = db_directory_update(&di);
+
+  free_di(&di, 1);
+
+  if (ret < 0 || id <= 0)
+  {
+    DPRINTF(E_LOG, L_DB, "Insert or update of directory failed '%s'\n", virtual_path);
+    return -1;
+  }
+
+  return id;
+}
+
 static void
 scheduled_cb(int fd, short what, void *arg)
 {
@@ -230,7 +269,7 @@ library_callback_schedule(library_cb cb, void *arg, struct timeval *wait, enum l
 
   event_add(cbreg->ev, wait);
 
-  DPRINTF(E_DBG, L_LIB, "Added library callback to %p (id %d), wait %ld.%06ld\n", cbreg->cb, idx_available, wait->tv_sec, wait->tv_usec);
+  DPRINTF(E_DBG, L_LIB, "Added library callback to %p (id %d), wait %ld.%06ld\n", cbreg->cb, idx_available, (long) wait->tv_sec, (long) wait->tv_usec);
 
   return idx_available;
 }
@@ -259,20 +298,27 @@ handle_deferred_update_notifications(void)
 }
 
 static void
-purge_cruft(time_t start)
+purge_cruft(time_t start, enum scan_kind scan_kind)
 {
   DPRINTF(E_DBG, L_LIB, "Purging old library content\n");
-  db_purge_cruft(start);
+  if (scan_kind > 0)
+    db_purge_cruft_bysource(start, scan_kind);
+  else
+    db_purge_cruft(start);
   db_groups_cleanup();
   db_queue_cleanup();
 
-  DPRINTF(E_DBG, L_LIB, "Purging old artwork content\n");
-  cache_artwork_purge_cruft(start);
+  if (scan_kind <= 0)
+    {
+      DPRINTF(E_DBG, L_LIB, "Purging old artwork content\n");
+      cache_artwork_purge_cruft(start);
+    }
 }
 
 static enum command_state
 rescan(void *arg, int *ret)
 {
+  enum scan_kind *scan_kind;
   time_t starttime;
   time_t endtime;
   int i;
@@ -281,20 +327,29 @@ rescan(void *arg, int *ret)
   listener_notify(LISTENER_UPDATE);
   starttime = time(NULL);
 
+  scan_kind = arg;
+
   for (i = 0; sources[i]; i++)
     {
       if (!sources[i]->disabled && sources[i]->rescan)
 	{
-	  DPRINTF(E_INFO, L_LIB, "Rescan library source '%s'\n", sources[i]->name);
-	  sources[i]->rescan();
+	  if (*scan_kind > 0 && *scan_kind != sources[i]->scan_kind)
+	    {
+	      DPRINTF(E_DBG, L_LIB, "Skipping library source '%s'\n", db_scan_kind_label(sources[i]->scan_kind));
+	    }
+	  else
+	    {
+	      DPRINTF(E_INFO, L_LIB, "Rescan library source '%s'\n", db_scan_kind_label(sources[i]->scan_kind));
+	      sources[i]->rescan();
+	    }
 	}
       else
 	{
-	  DPRINTF(E_INFO, L_LIB, "Library source '%s' is disabled\n", sources[i]->name);
+	  DPRINTF(E_INFO, L_LIB, "Library source '%s' is disabled\n", db_scan_kind_label(sources[i]->scan_kind));
 	}
     }
 
-  purge_cruft(starttime);
+  purge_cruft(starttime, *scan_kind);
 
   DPRINTF(E_DBG, L_LIB, "Running post library scan jobs\n");
   db_hook_post_scan();
@@ -315,6 +370,7 @@ rescan(void *arg, int *ret)
 static enum command_state
 metarescan(void *arg, int *ret)
 {
+  enum scan_kind *scan_kind;
   time_t starttime;
   time_t endtime;
   int i;
@@ -323,20 +379,29 @@ metarescan(void *arg, int *ret)
   listener_notify(LISTENER_UPDATE);
   starttime = time(NULL);
 
+  scan_kind = arg;
+
   for (i = 0; sources[i]; i++)
     {
       if (!sources[i]->disabled && sources[i]->metarescan)
 	{
-	  DPRINTF(E_INFO, L_LIB, "Meta rescan library source '%s'\n", sources[i]->name);
-	  sources[i]->metarescan();
+	  if (*scan_kind > 0 && *scan_kind != sources[i]->scan_kind)
+	    {
+	      DPRINTF(E_DBG, L_LIB, "Skipping library source '%s'\n", db_scan_kind_label(sources[i]->scan_kind));
+	    }
+	  else
+	    {
+	      DPRINTF(E_INFO, L_LIB, "Meta rescan library source '%s'\n", db_scan_kind_label(sources[i]->scan_kind));
+	      sources[i]->metarescan();
+	    }
 	}
       else
 	{
-	  DPRINTF(E_INFO, L_LIB, "Library source '%s' is disabled\n", sources[i]->name);
+	  DPRINTF(E_INFO, L_LIB, "Library source '%s' is disabled\n", db_scan_kind_label(sources[i]->scan_kind));
 	}
     }
 
-  purge_cruft(starttime);
+  purge_cruft(starttime, *scan_kind);
 
   DPRINTF(E_DBG, L_LIB, "Running post library scan jobs\n");
   db_hook_post_scan();
@@ -374,12 +439,12 @@ fullrescan(void *arg, int *ret)
     {
       if (!sources[i]->disabled && sources[i]->fullrescan)
 	{
-	  DPRINTF(E_INFO, L_LIB, "Full-rescan library source '%s'\n", sources[i]->name);
+	  DPRINTF(E_INFO, L_LIB, "Full-rescan library source '%s'\n", db_scan_kind_label(sources[i]->scan_kind));
 	  sources[i]->fullrescan();
 	}
       else
 	{
-	  DPRINTF(E_INFO, L_LIB, "Library source '%s' is disabled\n", sources[i]->name);
+	  DPRINTF(E_INFO, L_LIB, "Library source '%s' is disabled\n", db_scan_kind_label(sources[i]->scan_kind));
 	}
     }
 
@@ -409,7 +474,7 @@ playlist_item_add(void *arg, int *retval)
     {
       if (sources[i]->disabled || !sources[i]->playlist_item_add)
 	{
-	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support playlist_item_add\n", sources[i]->name);
+	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support playlist_item_add\n", db_scan_kind_label(sources[i]->scan_kind));
 	  continue;
 	}
 
@@ -417,7 +482,7 @@ playlist_item_add(void *arg, int *retval)
 
       if (ret == LIBRARY_OK)
 	{
-	  DPRINTF(E_DBG, L_LIB, "Adding item '%s' to playlist '%s' with library source '%s'\n", param->vp_item, param->vp_playlist, sources[i]->name);
+	  DPRINTF(E_DBG, L_LIB, "Adding item '%s' to playlist '%s' with library source '%s'\n", param->vp_item, param->vp_playlist, db_scan_kind_label(sources[i]->scan_kind));
 	  listener_notify(LISTENER_STORED_PLAYLIST);
 	  break;
 	}
@@ -440,7 +505,7 @@ playlist_remove(void *arg, int *retval)
     {
       if (sources[i]->disabled || !sources[i]->playlist_remove)
 	{
-	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support playlist_remove\n", sources[i]->name);
+	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support playlist_remove\n", db_scan_kind_label(sources[i]->scan_kind));
 	  continue;
 	}
 
@@ -448,7 +513,7 @@ playlist_remove(void *arg, int *retval)
 
       if (ret == LIBRARY_OK)
 	{
-	  DPRINTF(E_DBG, L_LIB, "Removing playlist '%s' with library source '%s'\n", virtual_path, sources[i]->name);
+	  DPRINTF(E_DBG, L_LIB, "Removing playlist '%s' with library source '%s'\n", virtual_path, db_scan_kind_label(sources[i]->scan_kind));
 	  listener_notify(LISTENER_STORED_PLAYLIST);
 	  break;
 	}
@@ -472,7 +537,7 @@ queue_item_add(void *arg, int *retval)
     {
       if (sources[i]->disabled || !sources[i]->queue_item_add)
         {
-	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support queue_add\n", sources[i]->name);
+	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support queue_add\n", db_scan_kind_label(sources[i]->scan_kind));
 	  continue;
 	}
 
@@ -480,7 +545,7 @@ queue_item_add(void *arg, int *retval)
 
       if (ret == LIBRARY_OK)
 	{
-	  DPRINTF(E_DBG, L_LIB, "Items for path '%s' from library source '%s' added to the queue\n", param->path, sources[i]->name);
+	  DPRINTF(E_DBG, L_LIB, "Items for path '%s' from library source '%s' added to the queue\n", param->path, db_scan_kind_label(sources[i]->scan_kind));
 	  break;
 	}
     }
@@ -505,7 +570,7 @@ queue_save(void *arg, int *retval)
     {
       if (sources[i]->disabled || !sources[i]->queue_save)
 	{
-	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support queue_save\n", sources[i]->name);
+	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support queue_save\n", db_scan_kind_label(sources[i]->scan_kind));
 	  continue;
 	}
 
@@ -513,7 +578,7 @@ queue_save(void *arg, int *retval)
 
       if (ret == LIBRARY_OK)
 	{
-	  DPRINTF(E_DBG, L_LIB, "Saving queue to path '%s' with library source '%s'\n", virtual_path, sources[i]->name);
+	  DPRINTF(E_DBG, L_LIB, "Saving queue to path '%s' with library source '%s'\n", virtual_path, db_scan_kind_label(sources[i]->scan_kind));
 	  listener_notify(LISTENER_STORED_PLAYLIST);
 	  break;
 	}
@@ -526,25 +591,25 @@ queue_save(void *arg, int *retval)
 static enum command_state
 item_add(void *arg, int *retval)
 {
-  const char *path = arg;
+  struct item_param *param = arg;
   int i;
   int ret = LIBRARY_ERROR;
 
-  DPRINTF(E_DBG, L_LIB, "Adding item to library '%s'\n", path);
+  DPRINTF(E_DBG, L_LIB, "Adding item to library '%s'\n", param->path);
 
   for (i = 0; sources[i]; i++)
     {
       if (sources[i]->disabled || !sources[i]->item_add)
 	{
-	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support add_item\n", sources[i]->name);
+	  DPRINTF(E_DBG, L_LIB, "Library source '%s' is disabled or does not support add_item\n", db_scan_kind_label(sources[i]->scan_kind));
 	  continue;
 	}
 
-      ret = sources[i]->item_add(path);
+      ret = sources[i]->item_add(param->path);
 
       if (ret == LIBRARY_OK)
 	{
-	  DPRINTF(E_DBG, L_LIB, "Add item to path '%s' with library source '%s'\n", path, sources[i]->name);
+	  DPRINTF(E_DBG, L_LIB, "Add item to path '%s' with library source '%s'\n", param->path, db_scan_kind_label(sources[i]->scan_kind));
 	  listener_notify(LISTENER_DATABASE);
 	  break;
 	}
@@ -564,6 +629,93 @@ item_add(void *arg, int *retval)
   return COMMAND_END;
 }
 
+static int
+write_metadata(struct media_file_info *mfi)
+{
+  int ret;
+  int i;
+
+  for (i = 0; sources[i]; i++)
+    {
+      if (sources[i]->disabled || !sources[i]->write_metadata)
+	continue;
+
+      ret = sources[i]->write_metadata(mfi);
+      if (ret == LIBRARY_OK)
+	return ret;
+    }
+
+  return LIBRARY_PATH_INVALID;
+}
+
+static enum command_state
+item_attrib_save(void *arg, int *retval)
+{
+  struct item_param *param = arg;
+  struct media_file_info *mfi = NULL;
+  int ret;
+
+  if (scanning)
+    goto error;
+
+  mfi = db_file_fetch_byid(param->id);
+  if (!mfi)
+    goto error;
+
+  *retval = LIBRARY_OK;
+
+  switch (param->attrib)
+    {
+      case LIBRARY_ATTRIB_RATING:
+	if (param->value > DB_FILES_RATING_MAX)
+	  goto error;
+
+	mfi->rating = param->value;
+
+	if (cfg_getbool(cfg_getsec(cfg, "library"), "write_rating"))
+	  *retval = write_metadata(mfi);
+
+        listener_notify(LISTENER_RATING);
+	break;
+
+      case LIBRARY_ATTRIB_USERMARK:
+	mfi->usermark = param->value;
+	break;
+
+      case LIBRARY_ATTRIB_PLAY_COUNT:
+	mfi->play_count = param->value;
+	break;
+
+      case LIBRARY_ATTRIB_SKIP_COUNT:
+	mfi->skip_count = param->value;
+	break;
+
+      case LIBRARY_ATTRIB_TIME_PLAYED:
+	mfi->time_played = param->value;
+	break;
+
+      case LIBRARY_ATTRIB_TIME_SKIPPED:
+	mfi->time_skipped = param->value;
+	break;
+
+       default:
+	goto error;
+    }
+
+  ret = db_file_update(mfi);
+  if (ret < 0)
+    goto error;
+
+  free_mfi(mfi, 0);
+  return COMMAND_END;
+
+ error:
+  DPRINTF(E_LOG, L_LIB, "Error updating attribute %d to %d for file with id %d\n", param->attrib, param->value, param->id);
+  *retval = LIBRARY_ERROR;
+  free_mfi(mfi, 0);
+  return COMMAND_END;
+}
+
 // Callback to notify listeners of database changes
 static void
 update_trigger_cb(int fd, short what, void *arg)
@@ -597,29 +749,39 @@ update_trigger(void *arg, int *retval)
 /* ----------------------- LIBRARY EXTERNAL INTERFACE ---------------------- */
 
 void
-library_rescan()
+library_rescan(enum scan_kind scan_kind)
 {
+  int *param;
+
   if (scanning)
     {
       DPRINTF(E_INFO, L_LIB, "Scan already running, ignoring request to trigger a new init scan\n");
       return;
     }
 
-  scanning = true; // TODO Guard "scanning" with a mutex
-  commands_exec_async(cmdbase, rescan, NULL);
+  scanning = true;
+  param = malloc(sizeof(int));
+  *param = scan_kind;
+
+  commands_exec_async(cmdbase, rescan, param);
 }
 
 void
-library_metarescan()
+library_metarescan(enum scan_kind scan_kind)
 {
+  int *param;
+
   if (scanning)
     {
       DPRINTF(E_INFO, L_LIB, "Scan already running, ignoring request to trigger metadata scan\n");
       return;
     }
 
-  scanning = true; // TODO Guard "scanning" with a mutex
-  commands_exec_async(cmdbase, metarescan, NULL);
+  scanning = true;
+  param = malloc(sizeof(int));
+  *param = scan_kind;
+
+  commands_exec_async(cmdbase, metarescan, param);
 }
 
 void
@@ -631,7 +793,7 @@ library_fullrescan()
       return;
     }
 
-  scanning = true; // TODO Guard "scanning" with a mutex
+  scanning = true;
   commands_exec_async(cmdbase, fullrescan, NULL);
 }
 
@@ -648,7 +810,15 @@ initscan()
   listener_notify(LISTENER_UPDATE);
 
   // Only clear the queue if enabled (default) in config
-  clear_queue_disabled = cfg_getbool(cfg_getsec(cfg, "mpd"), "clear_queue_on_stop_disable");
+  clear_queue_disabled = cfg_getbool(cfg_getsec(cfg, "library"), "clear_queue_on_stop_disable");
+
+  /* Handle deprecated config options */
+  if (0 < cfg_opt_size(cfg_getopt(cfg_getsec(cfg, "mpd"), "clear_queue_on_stop_disable")))
+    {
+      DPRINTF(E_LOG, L_MPD, "Found deprecated option 'clear_queue_on_stop_disable' in section 'mpd', please update configuration file (move option to section 'library').\n");
+      clear_queue_disabled = cfg_getbool(cfg_getsec(cfg, "mpd"), "clear_queue_on_stop_disable");
+    }
+
   if (!clear_queue_disabled)
     {
       db_queue_clear(0);
@@ -662,7 +832,7 @@ initscan()
 
   if (! (cfg_getbool(cfg_getsec(cfg, "library"), "filescan_disable")))
     {
-      purge_cruft(starttime);
+      purge_cruft(starttime, 0);
 
       DPRINTF(E_DBG, L_LIB, "Running post library scan jobs\n");
       db_hook_post_scan();
@@ -771,6 +941,8 @@ int
 library_queue_item_add(const char *path, int position, char reshuffle, uint32_t item_id, int *count, int *new_item_id)
 {
   struct queue_item_add_param param;
+  int count_internal;
+  int new_item_id_internal;
 
   if (library_is_scanning())
     return -1;
@@ -779,8 +951,8 @@ library_queue_item_add(const char *path, int position, char reshuffle, uint32_t
   param.position = position;
   param.reshuffle = reshuffle;
   param.item_id = item_id;
-  param.count = count;
-  param.new_item_id = new_item_id;
+  param.count = count ? count : &count_internal;
+  param.new_item_id = new_item_id ? new_item_id : &new_item_id_internal;
 
   return commands_exec_sync(cmdbase, queue_item_add, NULL, &param);
 }
@@ -788,15 +960,38 @@ library_queue_item_add(const char *path, int position, char reshuffle, uint32_t
 int
 library_item_add(const char *path)
 {
+  struct item_param param;
+
   if (scanning)
     {
       DPRINTF(E_INFO, L_LIB, "Scan already running, ignoring request to add item '%s'\n", path);
       return -1;
     }
 
-  scanning = true; // TODO Guard "scanning" with a mutex
+  scanning = true;
 
-  return commands_exec_sync(cmdbase, item_add, NULL, (char *)path);
+  param.path = path;
+
+  return commands_exec_sync(cmdbase, item_add, NULL, &param);
+}
+
+void
+library_item_attrib_save(uint32_t id, enum library_attrib attrib, uint32_t value)
+{
+  struct item_param *param;
+
+  param = malloc(sizeof(struct item_param));
+  param->id = id;
+  param->attrib = attrib;
+  param->value = value;
+
+  commands_exec_async(cmdbase, item_attrib_save, param);
+}
+
+struct library_source **
+library_sources(void)
+{
+  return sources;
 }
 
 int
@@ -863,7 +1058,7 @@ library_init(void)
     {
       if (!sources[i]->initscan || !sources[i]->rescan || !sources[i]->metarescan || !sources[i]->fullrescan)
 	{
-	  DPRINTF(E_FATAL, L_LIB, "BUG: library source '%s' is missing a scanning method\n", sources[i]->name);
+	  DPRINTF(E_FATAL, L_LIB, "BUG: library source '%s' is missing a scanning method\n", db_scan_kind_label(sources[i]->scan_kind));
 	  return -1;
 	}
 
