@@ -58,6 +58,8 @@ playlist_type(const char *path)
 
   if (strcasecmp(ptr, ".m3u") == 0)
     return PLAYLIST_M3U;
+  else if (strcasecmp(ptr, ".m3u8") == 0)
+    return PLAYLIST_M3U;
   else if (strcasecmp(ptr, ".pls") == 0)
     return PLAYLIST_PLS;
   else if (strcasecmp(ptr, ".smartpl") == 0)
@@ -253,6 +255,7 @@ process_url(int pl_id, const char *path, struct media_file_info *mfi)
       m3u.album_artist = safe_strdup(mfi->album_artist);
       m3u.album = safe_strdup(mfi->album);
       m3u.title = safe_strdup(mfi->title);
+      m3u.genre = safe_strdup(mfi->genre);
 
       scan_metadata_stream(mfi, path);
 
@@ -264,6 +267,8 @@ process_url(int pl_id, const char *path, struct media_file_info *mfi)
 	swap_pointers(&mfi->album, &m3u.album);
       if (m3u.title)
 	swap_pointers(&mfi->title, &m3u.title);
+      if (m3u.genre)
+	swap_pointers(&mfi->genre, &m3u.genre);
 
       free_mfi(&m3u, 1);
     }
@@ -366,6 +371,7 @@ static int
 playlist_prepare(const char *path, time_t mtime)
 {
   struct playlist_info *pli;
+  char *old_title;
   int pl_id;
 
   pli = db_pl_fetch_bypath(path);
@@ -384,7 +390,16 @@ playlist_prepare(const char *path, time_t mtime)
       return pl_id;
     }
 
-  db_pl_ping(pli->id);
+  // So we already have the playlist, but maybe it has been renamed
+  old_title = pli->title;
+  pli->title = title_from_path(path);
+
+  if (strcasecmp(old_title, pli->title) != 0)
+    db_pl_update(pli);
+  else
+    db_pl_ping(pli->id);
+
+  free(old_title);
 
   // mtime == db_timestamp is also treated as a modification because some editors do
   // stuff like 1) close the file with no changes (leading us to update db_timestamp),
@@ -451,6 +466,10 @@ scan_playlist(const char *file, time_t mtime, int dir_id)
     {
       len = strlen(buf);
 
+      // Check for and strip byte-order mark
+      if (memcmp("\xef\xbb\xbf", buf, 3) == 0)
+	memmove(buf, buf + 3, len - 3 + 1);
+
       // rtrim and check that length is sane (ignore blank lines)
       while ((len > 0) && isspace(buf[len - 1]))
 	{
@@ -471,11 +490,7 @@ scan_playlist(const char *file, time_t mtime, int dir_id)
       else if (pl_format == PLAYLIST_M3U)
 	path = buf;
 
-      if (!path)
-	continue;
-
-      // Check that first char is sane for a path
-      if ((!isalnum(path[0])) && (path[0] != '/') && (path[0] != '.'))
+      if (!path || path[0] == '\0' || path[0] == '#')
 	continue;
 
       // URLs and playlists will be added to library, tracks should already be there
