@@ -130,11 +130,16 @@ struct spotify_credentials
   time_t token_time_requested;
 };
 
+struct spotify_http_session
+{
+  pthread_mutex_t lock;
+  struct http_client_session session;
+};
+
+static struct spotify_http_session spotify_http_session = { .lock = PTHREAD_MUTEX_INITIALIZER };
+
 static struct spotify_credentials spotify_credentials;
-
-// Mutex to avoid conflicting requests for access tokens and protects accessing the credentials from different threads
-static pthread_mutex_t token_lck;
-
+static pthread_mutex_t spotify_credentials_lock = PTHREAD_MUTEX_INITIALIZER;
 
 // The base playlist id for all Spotify playlists in the db
 static int spotify_base_plid;
@@ -142,11 +147,12 @@ static int spotify_base_plid;
 // Flag to avoid triggering playlist change events while the (re)scan is running
 static bool scanning;
 
-
 // Endpoints and credentials for the web api
-static const char *spotify_client_id     = "0e684a5422384114a8ae7ac020f01789";
-static const char *spotify_client_secret = "232af95f39014c9ba218285a5c11a239";
-static const char *spotify_scope         = "playlist-read-private playlist-read-collaborative user-library-read user-read-private streaming";
+static const char *spotify_client_id;
+static const char *spotify_client_secret;
+static const char *spotify_redirect_uri;
+
+static const char *spotify_scope         = "playlist-read-collaborative playlist-read-private streaming user-follow-read user-library-read user-read-private";
 
 static const char *spotify_auth_uri      = "https://accounts.spotify.com/authorize";
 static const char *spotify_token_uri     = "https://accounts.spotify.com/api/token";
@@ -163,7 +169,6 @@ static const char *spotify_shows_uri           = "https://api.spotify.com/v1/me/
 static const char *spotify_shows_episodes_uri  = "https://api.spotify.com/v1/shows/%s/episodes";
 static const char *spotify_episode_uri         = "https://api.spotify.com/v1/episodes/%s";
 
-static struct http_client_session session = { 0 };
 
 static enum spotify_item_type
 parse_type_from_uri(const char *uri)
@@ -198,8 +203,173 @@ parse_type_from_uri(const char *uri)
 }
 
 static void
-free_credentials(void)
+credentials_update_token(const char *access_token, const char *refresh_token, const char *scope, int32_t expires_in)
 {
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_credentials_lock));
+
+  free(spotify_credentials.access_token);
+  free(spotify_credentials.refresh_token);
+  free(spotify_credentials.granted_scope);
+
+  spotify_credentials.access_token = safe_strdup(access_token);
+  spotify_credentials.refresh_token = safe_strdup(refresh_token);
+  spotify_credentials.granted_scope = safe_strdup(scope);
+  if (expires_in > 0)
+    spotify_credentials.token_expires_in = expires_in;
+  else
+    spotify_credentials.token_expires_in = 3600;
+  spotify_credentials.token_time_requested = time(NULL);
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_credentials_lock));
+}
+
+static void
+credentials_update_user(const char *user, const char *country)
+{
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_credentials_lock));
+
+  free(spotify_credentials.user);
+  free(spotify_credentials.user_country);
+
+  spotify_credentials.user = safe_strdup(user);
+  spotify_credentials.user_country = safe_strdup(country);
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_credentials_lock));
+}
+
+static void
+credentials_get_auth_header(char *header, size_t header_size)
+{
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_credentials_lock));
+
+  snprintf(header, header_size, "Authorization: Bearer %s", spotify_credentials.access_token);
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_credentials_lock));
+}
+
+static char *
+credentials_query_param_market(const char *href)
+{
+  char *next_href = NULL;
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_credentials_lock));
+
+  if (!spotify_credentials.user_country)
+    {
+      next_href = safe_strdup(href);
+    }
+  else
+    {
+      if (strchr(href, '?'))
+        next_href = safe_asprintf("%s&market=%s", href, spotify_credentials.user_country);
+      else
+        next_href = safe_asprintf("%s?market=%s", href, spotify_credentials.user_country);
+    }
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_credentials_lock));
+
+  return next_href;
+}
+
+static bool
+credentials_token_valid(void)
+{
+  bool valid = false;
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_credentials_lock));
+
+  if (spotify_credentials.access_token
+        && spotify_credentials.token_time_requested
+        && difftime(time(NULL), spotify_credentials.token_time_requested) < spotify_credentials.token_expires_in)
+    {
+      valid = true; // Spotify token still valid
+    }
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_credentials_lock));
+
+  return valid;
+}
+
+static bool
+credentials_token_exists(void)
+{
+  bool exists = false;
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_credentials_lock));
+
+  exists = (spotify_credentials.access_token != NULL);
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_credentials_lock));
+
+  return exists;
+}
+
+static void
+credentials_user_token_get(char **user, char **token)
+{
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_credentials_lock));
+
+  if (user)
+    *user = safe_strdup(spotify_credentials.user);
+  if (token)
+    *token = safe_strdup(spotify_credentials.access_token);
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_credentials_lock));
+}
+
+static void
+credentials_token_info(struct spotifywebapi_access_token *info)
+{
+  memset(info, 0, sizeof(struct spotifywebapi_access_token));
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_credentials_lock));
+  
+  if (spotify_credentials.token_time_requested > 0)
+    info->expires_in = spotify_credentials.token_expires_in - difftime(time(NULL), spotify_credentials.token_time_requested);
+  else
+    info->expires_in = 0;
+
+  info->token = safe_strdup(spotify_credentials.access_token);
+  
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_credentials_lock));
+}
+
+static void
+credentials_status_info(struct spotifywebapi_status_info *info)
+{
+  memset(info, 0, sizeof(struct spotifywebapi_status_info));
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_credentials_lock));
+
+  info->token_valid = (spotify_credentials.access_token != NULL);
+  if (spotify_credentials.user)
+    {
+      strncpy(info->user, spotify_credentials.user, (sizeof(info->user) - 1));
+    }
+  if (spotify_credentials.user_country)
+    {
+      strncpy(info->country, spotify_credentials.user_country, (sizeof(info->country) - 1));
+    }
+  if (spotify_credentials.granted_scope)
+    {
+      strncpy(info->granted_scope, spotify_credentials.granted_scope, (sizeof(info->granted_scope) - 1));
+    }
+  if (spotify_scope)
+    {
+      strncpy(info->required_scope, spotify_scope, (sizeof(info->required_scope) - 1));
+    }
+
+  info->client_id = spotify_client_id;
+  info->client_secret = spotify_client_secret;
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_credentials_lock));
+}
+
+static void
+credentials_clear()
+{
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_credentials_lock));
+
   free(spotify_credentials.access_token);
   free(spotify_credentials.refresh_token);
   free(spotify_credentials.granted_scope);
@@ -207,6 +377,8 @@ free_credentials(void)
   free(spotify_credentials.user);
 
   memset(&spotify_credentials, 0, sizeof(struct spotify_credentials));
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_credentials_lock));
 }
 
 static void
@@ -225,20 +397,17 @@ free_http_client_ctx(struct http_client_ctx *ctx)
   free(ctx);
 }
 
-static bool
-token_valid(void)
-{
-  return spotify_credentials.access_token != NULL;
-}
-
 static int
 request_access_tokens(struct keyval *kv, const char **err)
 {
   struct http_client_ctx ctx;
   char *param;
   char *body;
-  json_object *haystack;
-  const char *tmp;
+  json_object *haystack = NULL;
+  const char *access_token;
+  const char *refresh_token;
+  const char *scope;
+  uint32_t expires_in;
   int ret;
 
   param = http_form_urlencode(kv);
@@ -282,34 +451,8 @@ request_access_tokens(struct keyval *kv, const char **err)
       goto out_free_input_body;
     }
 
-  free(spotify_credentials.access_token);
-  spotify_credentials.access_token = NULL;
-
-  tmp = jparse_str_from_obj(haystack, "access_token");
-  if (tmp)
-    spotify_credentials.access_token = strdup(tmp);
-
-  tmp = jparse_str_from_obj(haystack, "refresh_token");
-  if (tmp)
-    {
-      free(spotify_credentials.refresh_token);
-      spotify_credentials.refresh_token = strdup(tmp);
-    }
-
-  tmp = jparse_str_from_obj(haystack, "scope");
-  if (tmp)
-    {
-      free(spotify_credentials.granted_scope);
-      spotify_credentials.granted_scope = strdup(tmp);
-    }
-
-  spotify_credentials.token_expires_in = jparse_int_from_obj(haystack, "expires_in");
-  if (spotify_credentials.token_expires_in == 0)
-    spotify_credentials.token_expires_in = 3600;
-
-  jparse_free(haystack);
-
-  if (!spotify_credentials.access_token)
+  access_token = jparse_str_from_obj(haystack, "access_token");
+  if (!access_token)
     {
       DPRINTF(E_LOG, L_SPOTIFY, "Could not find access token in reply: %s\n", body);
 
@@ -318,23 +461,28 @@ request_access_tokens(struct keyval *kv, const char **err)
       goto out_free_input_body;
     }
 
-  spotify_credentials.token_time_requested = time(NULL);
+  refresh_token = jparse_str_from_obj(haystack, "refresh_token");
+  if (refresh_token)
+    db_admin_set(DB_ADMIN_SPOTIFY_REFRESH_TOKEN, refresh_token);
 
-  if (spotify_credentials.refresh_token)
-    db_admin_set(DB_ADMIN_SPOTIFY_REFRESH_TOKEN, spotify_credentials.refresh_token);
+  scope = jparse_str_from_obj(haystack, "scope");
+  expires_in = jparse_int_from_obj(haystack, "expires_in");
+
+  credentials_update_token(access_token, refresh_token, scope, expires_in);
 
   ret = 0;
 
  out_free_input_body:
-  evbuffer_free(ctx.input_body);
-  free(param);
+   jparse_free(haystack);
+   evbuffer_free(ctx.input_body);
+   free(param);
  out_clear_kv:
 
   return ret;
 }
 
 /*
- * Request the api endpoint at 'href' and retuns the response body as
+ * Request the api endpoint at 'href' and returns the response body as
  * an allocated JSON object (must be freed by the caller) or NULL.
  *
  * @param href The spotify endpoint uri
@@ -355,16 +503,18 @@ request_endpoint(const char *uri)
 
   ctx->url = uri;
 
-  snprintf(bearer_token, sizeof(bearer_token), "Bearer %s", spotify_credentials.access_token);
+  credentials_get_auth_header(bearer_token, sizeof(bearer_token));
   if (keyval_add(ctx->output_headers, "Authorization", bearer_token) < 0)
     {
       DPRINTF(E_LOG, L_SPOTIFY, "Add bearer_token to keyval failed for request '%s'\n", uri);
       goto out;
     }
 
-  DPRINTF(E_DBG, L_SPOTIFY, "Request Spotify API endpoint: '%s')\n", uri);
+  DPRINTF(E_DBG, L_SPOTIFY, "Making request to '%s'\n", uri);
 
-  ret = http_client_request(ctx, &session);
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_http_session.lock));
+  ret = http_client_request(ctx, &spotify_http_session.session);
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_http_session.lock));
   if (ret < 0)
     {
       DPRINTF(E_LOG, L_SPOTIFY, "Request for '%s' failed\n", uri);
@@ -381,13 +531,13 @@ request_endpoint(const char *uri)
       goto out;
     }
 
-//  DPRINTF(E_DBG, L_SPOTIFY, "Wep api response for '%s'\n%s\n", uri, response_body);
+//  DPRINTF(E_DBG, L_SPOTIFY, "Web api response for '%s'\n%s\n", uri, response_body);
 
   json_response = json_tokener_parse(response_body);
   if (!json_response)
     DPRINTF(E_LOG, L_SPOTIFY, "JSON parser returned an error for '%s'\n", uri);
   else
-    DPRINTF(E_DBG, L_SPOTIFY, "Spotify API endpoint request: '%s'\n", uri);
+    DPRINTF(E_DBG, L_SPOTIFY, "Got JSON response for request to '%s'\n", uri);
 
  out:
   free_http_client_ctx(ctx);
@@ -401,25 +551,22 @@ request_endpoint(const char *uri)
  * API endpoint: https://api.spotify.com/v1/me
  */
 static int
-request_user_info(void)
+request_user_info()
 {
   json_object *response;
-
-  free(spotify_credentials.user_country);
-  spotify_credentials.user_country = NULL;
-  free(spotify_credentials.user);
-  spotify_credentials.user = NULL;
+  const char *user = NULL;
+  const char *user_country = NULL;
 
   response = request_endpoint(spotify_me_uri);
-
   if (response)
     {
-      spotify_credentials.user = safe_strdup(jparse_str_from_obj(response, "id"));
-      spotify_credentials.user_country = safe_strdup(jparse_str_from_obj(response, "country"));
+      user = jparse_str_from_obj(response, "id");
+      user_country = jparse_str_from_obj(response, "country");
+
+      DPRINTF(E_DBG, L_SPOTIFY, "User '%s', country '%s'\n", user, user_country);
+      credentials_update_user(user, user_country);
 
       jparse_free(response);
-
-      DPRINTF(E_DBG, L_SPOTIFY, "User '%s', country '%s'\n", spotify_credentials.user, spotify_credentials.user_country);
     }
 
   return 0;
@@ -431,20 +578,17 @@ request_user_info(void)
  * @return 0 on success, -1 on failure
  */
 static int
-token_get(const char *code, const char *redirect_uri, const char **err)
+token_get(const char *code, const char **err)
 {
-  struct keyval kv;
+  struct keyval kv = { 0 };
   int ret;
 
-  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&token_lck));
-
   *err = "";
-  memset(&kv, 0, sizeof(struct keyval));
   ret = ( (keyval_add(&kv, "grant_type", "authorization_code") == 0) &&
           (keyval_add(&kv, "code", code) == 0) &&
           (keyval_add(&kv, "client_id", spotify_client_id) == 0) &&
           (keyval_add(&kv, "client_secret", spotify_client_secret) == 0) &&
-          (keyval_add(&kv, "redirect_uri", redirect_uri) == 0) );
+          (keyval_add(&kv, "redirect_uri", spotify_redirect_uri) == 0) );
 
   if (!ret)
     {
@@ -459,8 +603,6 @@ token_get(const char *code, const char *redirect_uri, const char **err)
   if (ret == 0)
     request_user_info();
 
-  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&token_lck));
-
   return ret;
 }
 
@@ -476,28 +618,20 @@ token_get(const char *code, const char *redirect_uri, const char **err)
 static int
 token_refresh(void)
 {
-  struct keyval kv;
+  struct keyval kv = { 0 };
   char *refresh_token = NULL;
   const char *err;
   int ret;
 
-  memset(&kv, 0, sizeof(struct keyval));
-
-  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&token_lck));
-
-  if (spotify_credentials.token_time_requested && difftime(time(NULL), spotify_credentials.token_time_requested) < spotify_credentials.token_expires_in)
+  if (credentials_token_valid())
     {
-      DPRINTF(E_DBG, L_SPOTIFY, "Spotify token still valid\n");
-
-      CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&token_lck));
-      return 0;
+      return 0; // Spotify token still valid
     }
 
   ret = db_admin_get(&refresh_token, DB_ADMIN_SPOTIFY_REFRESH_TOKEN);
   if (ret < 0)
     {
-      DPRINTF(E_LOG, L_SPOTIFY, "No spotify refresh token found\n");
-      goto error;
+      return -1; // No refresh token (user not logged in)
     }
 
   DPRINTF(E_DBG, L_SPOTIFY, "Spotify refresh-token: '%s'\n", refresh_token);
@@ -513,23 +647,22 @@ token_refresh(void)
     }
 
   ret = request_access_tokens(&kv, &err);
+  if (ret < 0)
+    {
+      DPRINTF(E_LOG, L_SPOTIFY, "Error requesting access token: %s", err);
+      goto error;
+    }
 
-  if (ret == 0)
-    request_user_info();
+  request_user_info();
 
   free(refresh_token);
   keyval_clear(&kv);
 
-  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&token_lck));
-
-  return ret;
+  return 0;
 
  error:
   free(refresh_token);
   keyval_clear(&kv);
-
-  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&token_lck));
-
   return -1;
 }
 
@@ -541,6 +674,7 @@ token_refresh(void)
  * is checked and if necessary a token refresh request is issued before
  * requesting the given endpoint.
  *
+ * @param credentials Updated credentials
  * @param href The spotify endpoint uri
  * @return Response as JSON object or NULL
  */
@@ -582,11 +716,13 @@ typedef int (*paging_item_cb)(json_object *item, int index, int total, enum spot
  * @param pre_request_cb Callback function invoked before each request (optional)
  * @param post_request_cb Callback function invoked after each request (optional)
  * @param with_market If TRUE appends the user country as market to the request (applies track relinking)
+ * @param Spotify credentials
  * @param arg User data passed to each callback
  * @return 0 on success, -1 on failure
  */
 static int
-request_pagingobject_endpoint(const char *href, paging_item_cb item_cb, paging_request_cb pre_request_cb, paging_request_cb post_request_cb, bool with_market, enum spotify_request_type request_type, void *arg)
+request_pagingobject_endpoint(const char *href, paging_item_cb item_cb, paging_request_cb pre_request_cb, paging_request_cb post_request_cb,
+                              bool with_market, enum spotify_request_type request_type, void *arg)
 {
   char *next_href;
   json_object *response;
@@ -598,17 +734,10 @@ request_pagingobject_endpoint(const char *href, paging_item_cb item_cb, paging_r
   int total;
   int ret;
 
-  if (!with_market || !spotify_credentials.user_country)
-    {
-      next_href = safe_strdup(href);
-    }
+  if (!with_market)
+    next_href = safe_strdup(href);
   else
-    {
-      if (strchr(href, '?'))
-	next_href = safe_asprintf("%s&market=%s", href, spotify_credentials.user_country);
-      else
-	next_href = safe_asprintf("%s?market=%s", href, spotify_credentials.user_country);
-    }
+    next_href = credentials_query_param_market(href);
 
   while (next_href)
     {
@@ -641,8 +770,9 @@ request_pagingobject_endpoint(const char *href, paging_item_cb item_cb, paging_r
 	    {
 	      item = json_object_array_get_idx(items, i);
 	      if (!item)
-	        {
-		  DPRINTF(E_LOG, L_SPOTIFY, "Unexpected JSON: no item at index %d in '%s' (API endpoint: '%s')\n",
+		{
+		  // Occasionally Spotify returns a json array with null elements
+		  DPRINTF(E_DBG, L_SPOTIFY, "No item at index %d in '%s' (API endpoint: '%s')\n",
 			  i, json_object_to_json_string(items), href);
 		  continue;
 		}
@@ -650,7 +780,7 @@ request_pagingobject_endpoint(const char *href, paging_item_cb item_cb, paging_r
 	      ret = item_cb(item, (i + offset), total, request_type, arg);
 	      if (ret < 0)
 		{
-		  DPRINTF(E_LOG, L_SPOTIFY, "Unexpected JSON: error processing item at index %d '%s' (API endpoint: '%s')\n",
+		  DPRINTF(E_LOG, L_SPOTIFY, "Couldn't add item at index %d '%s' (API endpoint: '%s')\n",
 			  i, json_object_to_json_string(item), href);
 		}
 	    }
@@ -672,9 +802,10 @@ get_album_image(json_object *jsonalbum, int max_w)
   json_object *jsonimage;
   int image_count;
   int index;
-  const char *artwork_url;
-
-  artwork_url = NULL;
+  int width;
+  int candidate_width;
+  const char *artwork_url = NULL;
+  bool use_image;
 
   if (!json_object_object_get_ex(jsonalbum, "images", &jsonimages))
     {
@@ -682,24 +813,36 @@ get_album_image(json_object *jsonalbum, int max_w)
       return NULL;
     }
 
-  // Find first image that has a smaller width than the given max_w
-  // (this should avoid the need for resizing and improve performance at the cost of some quality loss)
-  // Note that Spotify returns the images ordered descending by width (widest image first)
-  // Special case is if no max width (max_w = 0) is given, the widest images will be used
+  // Find first image that has a smaller width than the given max_w (this should
+  // avoid the need for resizing and improve performance at the cost of some
+  // quality loss). If no sufficiently small image available, return smallest
+  // best alternative. Special case is if no max width (max_w = 0) is given, the
+  // widest images will be used.
+  //
+  // Note that Spotify should return the images ordered descending by width
+  // (widest image first), but at one point had a bug that meant they didn't, so
+  // we don't rely on that here.
   image_count = json_object_array_length(jsonimages);
-  for (index = 0; index < image_count; index++)
+  for (index = 0, candidate_width = 0; index < image_count; index++)
     {
       jsonimage = json_object_array_get_idx(jsonimages, index);
-      if (jsonimage)
-	{
-	  artwork_url = jparse_str_from_obj(jsonimage, "url");
+      if (!jsonimage)
+	continue;
 
-	  if (max_w <= 0  || jparse_int_from_obj(jsonimage, "width") <= max_w)
-	    {
-	      // We have the first image that has a smaller width than the given max_w
-	      break;
-	    }
-	}
+      width = jparse_int_from_obj(jsonimage, "width");
+
+      if (max_w == 0 || candidate_width == 0)
+	use_image = (width > candidate_width);
+      else if (candidate_width > max_w)
+	use_image = (width < candidate_width);
+      else
+	use_image = (candidate_width < width && width <= max_w);
+
+      if (!use_image)
+	continue;
+
+      candidate_width = width;
+      artwork_url = jparse_str_from_obj(jsonimage, "url");
     }
 
   return artwork_url;
@@ -1042,81 +1185,6 @@ request_episode(const char *path)
   return response;
 }
 
-/* Thread: httpd */
-char *
-spotifywebapi_oauth_uri_get(const char *redirect_uri)
-{
-  struct keyval kv;
-  char *param;
-  char *uri;
-  int uri_len;
-  int ret;
-
-  uri = NULL;
-  memset(&kv, 0, sizeof(struct keyval));
-  ret = ( (keyval_add(&kv, "client_id", spotify_client_id) == 0) &&
-	  (keyval_add(&kv, "response_type", "code") == 0) &&
-	  (keyval_add(&kv, "redirect_uri", redirect_uri) == 0) &&
-	  (keyval_add(&kv, "scope", spotify_scope) == 0) &&
-	  (keyval_add(&kv, "show_dialog", "false") == 0) );
-  if (!ret)
-    {
-      DPRINTF(E_LOG, L_SPOTIFY, "Cannot display Spotify oath interface (error adding parameters to keyval)\n");
-      goto out_clear_kv;
-    }
-
-  param = http_form_urlencode(&kv);
-  if (param)
-    {
-      uri_len = strlen(spotify_auth_uri) + strlen(param) + 3;
-
-      CHECK_NULL(L_SPOTIFY, uri = calloc(uri_len, sizeof(char)));
-
-      snprintf(uri, uri_len, "%s/?%s", spotify_auth_uri, param);
-
-      free(param);
-    }
-
- out_clear_kv:
-  keyval_clear(&kv);
-
-  return uri;
-}
-
-/* Thread: httpd */
-int
-spotifywebapi_oauth_callback(struct evkeyvalq *param, const char *redirect_uri, const char **errmsg)
-{
-  const char *code;
-  int ret;
-
-  *errmsg = NULL;
-
-  code = evhttp_find_header(param, "code");
-  if (!code)
-    {
-      *errmsg = "Error: Didn't receive a code from Spotify";
-      return -1;
-    }
-
-  DPRINTF(E_DBG, L_SPOTIFY, "Received OAuth code: %s\n", code);
-
-  ret = token_get(code, redirect_uri, errmsg);
-  if (ret < 0)
-    return -1;
-
-  ret = spotify_login_token(spotify_credentials.user, spotify_credentials.access_token, errmsg);
-  if (ret < 0)
-    return -1;
-
-  // Trigger scan after successful access to spotifywebapi
-  spotifywebapi_fullrescan();
-
-  listener_notify(LISTENER_SPOTIFY);
-
-  return 0;
-}
-
 static int
 transaction_start(void *arg)
 {
@@ -1169,7 +1237,7 @@ map_track_to_queueitem(struct db_queue_item *item, const struct spotify_track *t
 }
 
 static int
-queue_add_track(const char *uri, int position, char reshuffle, uint32_t item_id, int *count, int *new_item_id)
+queue_add_track(int *count, int *new_item_id, const char *uri, int position, char reshuffle, uint32_t item_id)
 {
   json_object *response = NULL;
   struct spotify_track track;
@@ -1244,7 +1312,7 @@ queue_add_album_tracks(json_object *item, int index, int total, enum spotify_req
 }
 
 static int
-queue_add_album(const char *uri, int position, char reshuffle, uint32_t item_id, int *count, int *new_item_id)
+queue_add_album(int *count, int *new_item_id, const char *uri, int position, char reshuffle, uint32_t item_id)
 {
   char *album_endpoint_uri = NULL;
   char *endpoint_uri = NULL;
@@ -1303,7 +1371,7 @@ queue_add_albums(json_object *item, int index, int total, enum spotify_request_t
 }
 
 static int
-queue_add_artist(const char *uri, int position, char reshuffle, uint32_t item_id, int *count, int *new_item_id)
+queue_add_artist(int *count, int *new_item_id, const char *uri, int position, char reshuffle, uint32_t item_id)
 {
   struct db_queue_add_info queue_add_info;
   char *endpoint_uri = NULL;
@@ -1365,7 +1433,7 @@ queue_add_playlist_tracks(json_object *item, int index, int total, enum spotify_
 }
 
 static int
-queue_add_playlist(const char *uri, int position, char reshuffle, uint32_t item_id, int *count, int *new_item_id)
+queue_add_playlist(int *count, int *new_item_id, const char *uri, int position, char reshuffle, uint32_t item_id)
 {
   char *endpoint_uri = NULL;
   struct db_queue_add_info queue_add_info;
@@ -1391,36 +1459,6 @@ queue_add_playlist(const char *uri, int position, char reshuffle, uint32_t item_
   return ret;
 }
 
-static int
-queue_item_add(const char *uri, int position, char reshuffle, uint32_t item_id, int *count, int *new_item_id)
-{
-  enum spotify_item_type type;
-
-  type = parse_type_from_uri(uri);
-  if (type == SPOTIFY_ITEM_TYPE_TRACK)
-    {
-      queue_add_track(uri, position, reshuffle, item_id, count, new_item_id);
-      return LIBRARY_OK;
-    }
-  else if (type == SPOTIFY_ITEM_TYPE_ARTIST)
-    {
-      queue_add_artist(uri, position, reshuffle, item_id, count, new_item_id);
-      return LIBRARY_OK;
-    }
-  else if (type == SPOTIFY_ITEM_TYPE_ALBUM)
-    {
-      queue_add_album(uri, position, reshuffle, item_id, count, new_item_id);
-      return LIBRARY_OK;
-    }
-  else if (type == SPOTIFY_ITEM_TYPE_PLAYLIST)
-    {
-      queue_add_playlist(uri, position, reshuffle, item_id, count, new_item_id);
-      return LIBRARY_OK;
-    }
-
-  return LIBRARY_PATH_INVALID;
-}
-
 
 /*
  * Returns the directory id for /spotify:/<artist>/<album>, if the directory (or the parent
@@ -1737,9 +1775,8 @@ scan_saved_shows(enum spotify_request_type request_type)
   return ret;
 }
 
-
 /*
- * Add a saved playlist tracks to the library
+ * Add a saved playlist's tracks to the library
  */
 static int
 saved_playlist_tracks_add(json_object *item, int index, int total, enum spotify_request_type request_type, void *arg)
@@ -1835,8 +1872,8 @@ saved_playlist_add(json_object *item, int index, int total, enum spotify_request
 
   if (!playlist.uri || !playlist.name || playlist.tracks_count == 0)
     {
-      DPRINTF(E_LOG, L_SPOTIFY, "Ignoring playlist '%s' with %d tracks (%s)\n", playlist.name, playlist.tracks_count, playlist.uri);
-      return -1;
+      DPRINTF(E_INFO, L_SPOTIFY, "Ignoring playlist '%s' with %d tracks (%s)\n", playlist.name, playlist.tracks_count, playlist.uri);
+      return 0; // Ignore
     }
 
   map_playlist_to_pli(&pli, &playlist);
@@ -1911,7 +1948,7 @@ scan(enum spotify_request_type request_type)
   time_t start;
   time_t end;
 
-  if (!token_valid() || scanning)
+  if (!credentials_token_exists() || scanning)
     {
       DPRINTF(E_DBG, L_SPOTIFY, "No valid web api token or scan already in progress, rescan ignored\n");
       return;
@@ -1922,6 +1959,7 @@ scan(enum spotify_request_type request_type)
 
   db_directory_enable_bypath("/spotify:");
   create_base_playlist();
+
   scan_saved_albums(request_type);
   scan_playlists(request_type);
   spotify_status_get(&sp_status);
@@ -1934,9 +1972,45 @@ scan(enum spotify_request_type request_type)
   DPRINTF(E_LOG, L_SPOTIFY, "Spotify scan completed in %.f sec\n", difftime(end, start));
 }
 
-/* Thread: library */
+
+/* --------------------------- Library interface ---------------------------- */
+/*                              Thread: library                               */
+
 static int
-initscan(void)
+spotifywebapi_library_queue_item_add(const char *uri, int position, char reshuffle, uint32_t item_id, int *count, int *new_item_id)
+{
+  enum spotify_item_type type;
+
+  type = parse_type_from_uri(uri);
+  if (type == SPOTIFY_ITEM_TYPE_TRACK)
+    {
+      queue_add_track(count, new_item_id, uri, position, reshuffle, item_id);
+      goto out;
+    }
+  else if (type == SPOTIFY_ITEM_TYPE_ARTIST)
+    {
+      queue_add_artist(count, new_item_id, uri, position, reshuffle, item_id);
+      goto out;
+    }
+  else if (type == SPOTIFY_ITEM_TYPE_ALBUM)
+    {
+      queue_add_album(count, new_item_id, uri, position, reshuffle, item_id);
+      goto out;
+    }
+  else if (type == SPOTIFY_ITEM_TYPE_PLAYLIST)
+    {
+      queue_add_playlist(count, new_item_id, uri, position, reshuffle, item_id);
+      goto out;
+    }
+
+  return LIBRARY_PATH_INVALID;
+
+ out:
+  return LIBRARY_OK;
+}
+
+static int
+spotifywebapi_library_initscan(void)
 {
   int ret;
 
@@ -1944,12 +2018,8 @@ initscan(void)
   ret = token_refresh();
   if (ret < 0)
     {
-      DPRINTF(E_LOG, L_SPOTIFY, "Spotify webapi token refresh failed. "
-	"In order to use Spotify, authorize the server to access your saved "
-	"tracks by visiting http://owntone.local:3689\n");
-
+      // User not logged in or error refreshing token
       db_spotify_purge();
-
       return 0;
     }
 
@@ -1964,7 +2034,6 @@ initscan(void)
 	"provide valid credentials by visiting http://owntone.local:3689\n");
 
       db_spotify_purge();
-
       return 0;
     }
 
@@ -1972,56 +2041,98 @@ initscan(void)
    * Scan saved tracks from the web api
    */
   scan(SPOTIFY_REQUEST_TYPE_RESCAN);
-
   return 0;
 }
 
-/* Thread: library */
 static int
-rescan(void)
+spotifywebapi_library_rescan(void)
 {
   scan(SPOTIFY_REQUEST_TYPE_RESCAN);
   return 0;
 }
 
-/* Thread: library */
 static int
-metarescan(void)
+spotifywebapi_library_metarescan(void)
 {
   scan(SPOTIFY_REQUEST_TYPE_METARESCAN);
   return 0;
 }
 
-/* Thread: library */
 static int
-fullrescan(void)
+spotifywebapi_library_fullrescan(void)
 {
   db_spotify_purge();
+
   scan(SPOTIFY_REQUEST_TYPE_RESCAN);
   return 0;
 }
 
-/* Thread: library */
+static int
+spotifywebapi_library_init()
+{
+  int ret;
+
+  spotify_client_id = cfg_getstr(cfg_getsec(cfg, "spotify"), "webapi_client_id");
+  spotify_client_secret = cfg_getstr(cfg_getsec(cfg, "spotify"), "webapi_client_secret");
+  spotify_redirect_uri = cfg_getstr(cfg_getsec(cfg, "spotify"), "redirect_uri");
+
+  ret = spotify_init();
+  if (ret < 0)
+    return -1;
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_http_session.lock));
+  http_client_session_init(&spotify_http_session.session);
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_http_session.lock));
+  return 0;
+}
+
+static void
+spotifywebapi_library_deinit()
+{
+  spotify_deinit();
+
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&spotify_http_session.lock));
+  http_client_session_deinit(&spotify_http_session.session);
+  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&spotify_http_session.lock));
+
+  credentials_clear();
+}
+
+struct library_source spotifyscanner =
+{
+  .scan_kind = SCAN_KIND_SPOTIFY,
+  .disabled = 0,
+  .queue_item_add = spotifywebapi_library_queue_item_add,
+  .initscan = spotifywebapi_library_initscan,
+  .rescan = spotifywebapi_library_rescan,
+  .metarescan = spotifywebapi_library_metarescan,
+  .fullrescan = spotifywebapi_library_fullrescan,
+  .init = spotifywebapi_library_init,
+  .deinit = spotifywebapi_library_deinit,
+};
+
+
+/* ------------------------ Public API command callbacks -------------------- */
+/*                              Thread: library                               */
+
 static enum command_state
 webapi_fullrescan(void *arg, int *ret)
 {
-  *ret = fullrescan();
+  *ret = spotifywebapi_library_fullrescan();
   return COMMAND_END;
 }
 
-/* Thread: library */
 static enum command_state
 webapi_rescan(void *arg, int *ret)
 {
-  *ret = rescan();
+  *ret = spotifywebapi_library_rescan();
   return COMMAND_END;
 }
 
-/* Thread: library */
 static enum command_state
 webapi_purge(void *arg, int *ret)
 {
-  free_credentials();
+  credentials_clear();
 
   db_spotify_purge();
   db_admin_delete(DB_ADMIN_SPOTIFY_REFRESH_TOKEN);
@@ -2030,6 +2141,92 @@ webapi_purge(void *arg, int *ret)
   return COMMAND_END;
 }
 
+
+/* ------------------------------ Public API -------------------------------- */
+
+char *
+spotifywebapi_oauth_uri_get(void)
+{
+  struct keyval kv = { 0 };
+  char *param;
+  char *uri;
+  int uri_len;
+  int ret;
+
+  uri = NULL;
+  ret = ( (keyval_add(&kv, "client_id", spotify_client_id) == 0) &&
+	  (keyval_add(&kv, "response_type", "code") == 0) &&
+	  (keyval_add(&kv, "redirect_uri", spotify_redirect_uri) == 0) &&
+	  (keyval_add(&kv, "scope", spotify_scope) == 0) &&
+	  (keyval_add(&kv, "show_dialog", "false") == 0) );
+  if (!ret)
+    {
+      DPRINTF(E_LOG, L_SPOTIFY, "Cannot display Spotify oath interface (error adding parameters to keyval)\n");
+      goto out_clear_kv;
+    }
+
+  param = http_form_urlencode(&kv);
+  if (param)
+    {
+      uri_len = strlen(spotify_auth_uri) + strlen(param) + 3;
+
+      CHECK_NULL(L_SPOTIFY, uri = calloc(uri_len, sizeof(char)));
+
+      snprintf(uri, uri_len, "%s/?%s", spotify_auth_uri, param);
+
+      free(param);
+    }
+
+ out_clear_kv:
+  keyval_clear(&kv);
+
+  return uri;
+}
+
+int
+spotifywebapi_oauth_callback(struct evkeyvalq *param, const char **errmsg)
+{
+  const char *code;
+  char *user = NULL;
+  char *access_token = NULL;
+  int ret;
+
+  *errmsg = NULL;
+
+  code = evhttp_find_header(param, "code");
+  if (!code)
+    {
+      *errmsg = "Error: Didn't receive a code from Spotify";
+      goto error;
+    }
+
+  DPRINTF(E_DBG, L_SPOTIFY, "Received OAuth code: %s\n", code);
+
+  ret = token_get(code, errmsg);
+  if (ret < 0)
+    goto error;
+
+  credentials_user_token_get(&user, &access_token);
+
+  ret = spotify_login(user, access_token, errmsg);
+  if (ret < 0)
+    goto error;
+
+  // Trigger scan after successful access to spotifywebapi
+  spotifywebapi_fullrescan();
+
+  listener_notify(LISTENER_SPOTIFY);
+
+  free(user);
+  free(access_token);
+  return 0;
+
+ error:
+  free(user);
+  free(access_token);
+  return -1;
+}
+
 void
 spotifywebapi_fullrescan(void)
 {
@@ -2091,85 +2288,12 @@ spotifywebapi_artwork_url_get(const char *uri, int max_w, int max_h)
 void
 spotifywebapi_status_info_get(struct spotifywebapi_status_info *info)
 {
-  memset(info, 0, sizeof(struct spotifywebapi_status_info));
-
-  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&token_lck));
-
-  info->token_valid = token_valid();
-  if (spotify_credentials.user)
-    {
-      strncpy(info->user, spotify_credentials.user, (sizeof(info->user) - 1));
-    }
-  if (spotify_credentials.user_country)
-    {
-      strncpy(info->country, spotify_credentials.user_country, (sizeof(info->country) - 1));
-    }
-  if (spotify_credentials.granted_scope)
-    {
-      strncpy(info->granted_scope, spotify_credentials.granted_scope, (sizeof(info->granted_scope) - 1));
-    }
-  if (spotify_scope)
-    {
-      strncpy(info->required_scope, spotify_scope, (sizeof(info->required_scope) - 1));
-    }
-
-  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&token_lck));
+  credentials_status_info(info);
 }
 
 void
 spotifywebapi_access_token_get(struct spotifywebapi_access_token *info)
 {
   token_refresh();
-
-  memset(info, 0, sizeof(struct spotifywebapi_access_token));
-
-  CHECK_ERR(L_SPOTIFY, pthread_mutex_lock(&token_lck));
-
-  if (spotify_credentials.token_time_requested > 0)
-    info->expires_in = spotify_credentials.token_expires_in - difftime(time(NULL), spotify_credentials.token_time_requested);
-  else
-    info->expires_in = 0;
-
-  info->token = safe_strdup(spotify_credentials.access_token);
-
-  CHECK_ERR(L_SPOTIFY, pthread_mutex_unlock(&token_lck));
+  credentials_token_info(info);
 }
-
-static int
-spotifywebapi_init()
-{
-  int ret;
-
-  CHECK_ERR(L_SPOTIFY, mutex_init(&token_lck));
-  ret = spotify_init();
-  if (ret < 0)
-    return -1;
-
-  http_client_session_init(&session);
-  return 0;
-}
-
-static void
-spotifywebapi_deinit()
-{
-  CHECK_ERR(L_SPOTIFY, pthread_mutex_destroy(&token_lck));
-
-  spotify_deinit();
-  http_client_session_deinit(&session);
-
-  free_credentials();
-}
-
-struct library_source spotifyscanner =
-{
-  .scan_kind = SCAN_KIND_SPOTIFY,
-  .disabled = 0,
-  .init = spotifywebapi_init,
-  .deinit = spotifywebapi_deinit,
-  .rescan = rescan,
-  .metarescan = metarescan,
-  .initscan = initscan,
-  .fullrescan = fullrescan,
-  .queue_item_add = queue_item_add,
-};
-
