@@ -47,9 +47,7 @@
 
 #include <getopt.h>
 #include <event2/event.h>
-#ifdef HAVE_LIBEVENT_PTHREADS
-# include <event2/thread.h>
-#endif
+#include <event2/thread.h>
 #include <libavutil/avutil.h>
 #include <libavutil/log.h>
 #include <libavformat/avformat.h>
@@ -58,7 +56,6 @@
 
 #include <pthread.h>
 #include <gcrypt.h>
-GCRY_THREAD_OPTION_PTHREAD_IMPL;
 
 #include "conffile.h"
 #include "db.h"
@@ -75,9 +72,11 @@ GCRY_THREAD_OPTION_PTHREAD_IMPL;
 #ifdef LASTFM
 # include "lastfm.h"
 #endif
+#include "listenbrainz.h"
 
-#define PIDFILE   STATEDIR "/run/" PACKAGE ".pid"
-#define WEB_ROOT  DATADIR "/htdocs"
+#define PIDFILE          STATEDIR "/run/" PACKAGE ".pid"
+#define WEB_ROOT         DATADIR "/htdocs"
+#define SQLITE_EXT_PATH  PKGLIBDIR "/" PACKAGE_NAME "-sqlext.so"
 
 struct event_base *evbase_main;
 
@@ -97,14 +96,19 @@ usage(char *program)
   printf("\n");
   printf("Usage: %s [options]\n\n", program);
   printf("Options:\n");
-  printf("  -d <number>    Log level (0-5)\n");
-  printf("  -D <dom,dom..> Log domains\n");
-  printf("  -c <file>      Use <file> as the configfile\n");
-  printf("  -P <file>      Write PID to specified file\n");
-  printf("  -f             Run in foreground\n");
-  printf("  -b <id>        ffid to be broadcast\n");
-  printf("  -v             Display version information\n");
-  printf("  -w <directory> Use <directory> as the web root directory for serving static files\n");
+  printf("  -d <number>     Log level (0-5)\n");
+  printf("  -D <dom,dom..>  Log domains\n");
+  printf("  -c <file>       Use <file> as the configuration file\n");
+  printf("  -P <file>       Write PID to specified file\n");
+  printf("  -f              Run in foreground\n");
+  printf("  -b <id>         ffid to be broadcast\n");
+  printf("  -v              Display version information\n");
+  printf("  -w <directory>  Use <directory> as the web root directory for serving static files\n");
+  printf("  --mdns-no-rsp   Don't announce RSP service via mDNS\n");
+  printf("  --mdns-no-daap  Don't announce DAAP service via mDNS\n");
+  printf("  --mdns-no-cname Don't register owntone.local as CNAME via mDNS\n");
+  printf("  --mdns-no-web   Don't announce web interface via mDNS\n");
+  printf("  -s <path>      Use <path> as the path for the OwnTone sqlite extension (owntone-sqlext.so)\n");
   printf("\n\n");
   printf("Available log domains:\n");
   logger_domains();
@@ -364,7 +368,7 @@ signal_signalfd_cb(int fd, short event, void *arg)
   struct signalfd_siginfo info;
   int status;
 
-  while (read(fd, &info, sizeof(struct signalfd_siginfo)) > 0)
+  while (read(fd, &info, sizeof(struct signalfd_siginfo)) == sizeof(struct signalfd_siginfo))
     {
       switch (info.ssi_signo)
 	{
@@ -480,19 +484,22 @@ int
 main(int argc, char **argv)
 {
   int option;
-  char *configfile;
-  bool background;
-  bool mdns_no_rsp;
-  bool mdns_no_daap;
-  bool mdns_no_cname;
-  bool mdns_no_web;
-  bool mdns_no_mpd;
-  int loglevel;
-  char *logdomains;
-  char *logfile;
-  char *ffid;
-  char *pidfile;
-  char *webroot;
+  char *configfile = CONFFILE;
+  bool background = true;
+  bool testrun = false;
+  bool mdns_no_rsp = false;
+  bool mdns_no_daap = false;
+  bool mdns_no_cname = false;
+  bool mdns_no_web = false;
+  bool mdns_no_mpd = false;
+  int loglevel = -1;
+  char *logdomains = NULL;
+  char *logfile = NULL;
+  char *logformat = NULL;
+  char *ffid = NULL;
+  char *pidfile = PIDFILE;
+  char *webroot = WEB_ROOT;
+  char *sqlite_extension_path = SQLITE_EXT_PATH;
   char **buildopts;
   const char *av_version;
   const char *gcry_version;
@@ -504,39 +511,28 @@ main(int argc, char **argv)
   int i;
   int ret;
 
-  struct option option_map[] =
-    {
-      { "ffid",         1, NULL, 'b' },
-      { "debug",        1, NULL, 'd' },
-      { "logdomains",   1, NULL, 'D' },
-      { "foreground",   0, NULL, 'f' },
-      { "config",       1, NULL, 'c' },
-      { "pidfile",      1, NULL, 'P' },
-      { "version",      0, NULL, 'v' },
-      { "webroot",      1, NULL, 'w' },
+  struct option option_map[] = {
+    { "ffid",          1, NULL, 'b' },
+    { "debug",         1, NULL, 'd' },
+    { "logdomains",    1, NULL, 'D' },
+    { "foreground",    0, NULL, 'f' },
+    { "config",        1, NULL, 'c' },
+    { "pidfile",       1, NULL, 'P' },
+    { "version",       0, NULL, 'v' },
+    { "webroot",       1, NULL, 'w' },
+    { "sqliteext",     1, NULL, 's' },
+    { "testrun",       0, NULL, 't' }, // Used for CI, not documented to user
 
-      { "mdns-no-rsp",  0, NULL, 512 },
-      { "mdns-no-daap", 0, NULL, 513 },
-      { "mdns-no-cname",0, NULL, 514 },
-      { "mdns-no-web",  0, NULL, 515 },
+    { "mdns-no-rsp",   0, NULL, 512 },
+    { "mdns-no-daap",  0, NULL, 513 },
+    { "mdns-no-cname", 0, NULL, 514 },
+    { "mdns-no-web",   0, NULL, 515 },
+    { "logformat",     1, NULL, 516 },
 
-      { NULL,           0, NULL, 0 }
-    };
+    { NULL,            0, NULL, 0   }
+  };
 
-  configfile = CONFFILE;
-  pidfile = PIDFILE;
-  webroot = WEB_ROOT;
-  loglevel = -1;
-  logdomains = NULL;
-  logfile = NULL;
-  background = true;
-  ffid = NULL;
-  mdns_no_rsp = false;
-  mdns_no_daap = false;
-  mdns_no_cname = false;
-  mdns_no_web = false;
-
-  while ((option = getopt_long(argc, argv, "D:d:c:P:fb:vw:", option_map, NULL)) != -1)
+  while ((option = getopt_long(argc, argv, "D:d:c:P:ftb:vw:s:", option_map, NULL)) != -1)
     {
       switch (option)
 	{
@@ -556,6 +552,14 @@ main(int argc, char **argv)
 	    mdns_no_web = true;
 	    break;
 
+	  case 516:
+	    logformat = optarg;
+	    break;
+
+	  case 't':
+	    testrun = true;
+	    break;
+
 	  case 'b':
 	    ffid = optarg;
 	    break;
@@ -593,6 +597,10 @@ main(int argc, char **argv)
 	    webroot = optarg;
 	    break;
 
+	  case 's':
+	    sqlite_extension_path = optarg;
+	    break;
+
 	  default:
 	    usage(argv[0]);
 	    return EXIT_FAILURE;
@@ -600,7 +608,7 @@ main(int argc, char **argv)
 	}
     }
 
-  ret = logger_init(NULL, NULL, (loglevel < 0) ? E_LOG : loglevel);
+  ret = logger_init(NULL, NULL, (loglevel < 0) ? E_LOG : loglevel, NULL);
   if (ret != 0)
     {
       fprintf(stderr, "Could not initialize log facility\n");
@@ -622,10 +630,11 @@ main(int argc, char **argv)
   /* Reinit log facility with configfile values */
   if (loglevel < 0)
     loglevel = cfg_getint(cfg_getsec(cfg, "general"), "loglevel");
-
+  if (!logformat)
+    logformat = cfg_getstr(cfg_getsec(cfg, "general"), "logformat");
   logfile = cfg_getstr(cfg_getsec(cfg, "general"), "logfile");
 
-  ret = logger_init(logfile, logdomains, loglevel);
+  ret = logger_init(logfile, logdomains, loglevel, logformat);
   if (ret != 0)
     {
       fprintf(stderr, "Could not reinitialize log facility with config file settings\n");
@@ -684,9 +693,6 @@ main(int argc, char **argv)
   /* Initialize libcurl */
   curl_global_init(CURL_GLOBAL_DEFAULT);
 
-  /* Initialize libgcrypt */
-  gcry_control(GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread);
-
   gcry_version = gcry_check_version(GCRYPT_VERSION);
   if (!gcry_version)
     {
@@ -734,9 +740,7 @@ main(int argc, char **argv)
   /* Initialize event base (after forking) */
   CHECK_NULL(L_MAIN, evbase_main = event_base_new());
 
-#ifdef HAVE_LIBEVENT_PTHREADS
   CHECK_ERR(L_MAIN, evthread_use_pthreads());
-#endif
 
   DPRINTF(E_LOG, L_MAIN, "mDNS init\n");
   ret = mdns_init();
@@ -750,7 +754,7 @@ main(int argc, char **argv)
 
   /* Initialize the database before starting */
   DPRINTF(E_INFO, L_MAIN, "Initializing database\n");
-  ret = db_init();
+  ret = db_init(sqlite_extension_path);
   if (ret < 0)
     {
       DPRINTF(E_FATAL, L_MAIN, "Database init failed\n");
@@ -829,7 +833,6 @@ main(int argc, char **argv)
       ret = EXIT_FAILURE;
       goto mpd_fail;
     }
-  mdns_no_mpd = false;
 #else
   mdns_no_mpd = true;
 #endif
@@ -837,6 +840,7 @@ main(int argc, char **argv)
 #ifdef LASTFM
   lastfm_init();
 #endif
+  listenbrainz_init();
 
   /* Start Remote pairing service */
   ret = remote_pairing_init();
@@ -911,7 +915,8 @@ main(int argc, char **argv)
   event_add(sig_event, NULL);
 
   /* Run the loop */
-  event_base_dispatch(evbase_main);
+  if (!testrun)
+    event_base_dispatch(evbase_main);
 
   DPRINTF(E_LOG, L_MAIN, "Stopping gracefully\n");
   ret = EXIT_SUCCESS;
