@@ -48,8 +48,13 @@
 # include <pthread_np.h>
 #endif
 
+#include <fcntl.h>
+#include <poll.h>
 #include <netdb.h>
 #include <arpa/inet.h>
+#include <ifaddrs.h> // getifaddrs
+
+#include <event2/http.h> // evhttp_bind
 
 #include <unistr.h>
 #include <uniconv.h>
@@ -120,20 +125,114 @@ static char *buildopts[] =
 
 /* ------------------------ Network utility functions ----------------------- */
 
+static bool
+prefix_is_equal(uint8_t *addr1, uint8_t *addr2, uint8_t *mask, size_t len)
+{
+  int i;
+
+  for (i = 0; i < len; i++)
+    {
+      if ((addr1[i] & mask[i]) != (addr2[i] & mask[i]))
+	return false;
+    }
+
+  return true;
+}
+
+// Checks if the address is in any of the interface subnets
+static bool
+net_address_is_local(union net_sockaddr *naddr)
+{
+  struct ifaddrs *ifaddrs;
+  struct ifaddrs *iap;
+  void *if_addr;
+  void *if_netmask;
+  void *addr;
+  int addr_len;
+  sa_family_t addr_family;
+  bool is_local = false;
+
+  // Point addr to the actual address data
+  if (naddr->sa.sa_family == AF_INET6 && IN6_IS_ADDR_V4MAPPED(&naddr->sin6.sin6_addr))
+    {
+      addr = naddr->sin6.sin6_addr.s6_addr + 12; // ipv4 over ipv6 prefix is 12 bytes
+      addr_len = sizeof(struct in6_addr) - 12;
+      addr_family = AF_INET;
+    }
+  else if (naddr->sa.sa_family == AF_INET6)
+    {
+      addr = naddr->sin6.sin6_addr.s6_addr;
+      addr_len = sizeof(struct in6_addr);
+      addr_family = AF_INET6;
+    }
+  else if (naddr->sa.sa_family == AF_INET)
+    {
+      addr = &naddr->sin.sin_addr.s_addr;
+      addr_len = sizeof(struct in_addr);
+      addr_family = AF_INET;
+    }
+  else
+    {
+      return false;
+    }
+
+  getifaddrs(&ifaddrs);
+
+  for (iap = ifaddrs; iap && !is_local; iap = iap->ifa_next)
+    {
+      if (!iap->ifa_addr || !iap->ifa_netmask)
+	continue;
+      if (iap->ifa_addr->sa_family != addr_family || iap->ifa_netmask->sa_family != addr_family)
+	continue;
+
+      if (addr_family == AF_INET6)
+	{
+	  if_addr = ((struct sockaddr_in6 *)iap->ifa_addr)->sin6_addr.s6_addr;
+	  if_netmask = ((struct sockaddr_in6 *)iap->ifa_netmask)->sin6_addr.s6_addr;
+	}
+      else
+	{
+	  if_addr = &((struct sockaddr_in *)iap->ifa_addr)->sin_addr.s_addr;
+	  if_netmask = &((struct sockaddr_in *)iap->ifa_netmask)->sin_addr.s_addr;
+	}
+
+      is_local = prefix_is_equal(addr, if_addr, if_netmask, addr_len);
+    }
+
+  freeifaddrs(ifaddrs);
+
+  return is_local;
+}
+
+static bool
+net_address_has_prefix(union net_sockaddr *naddr, const char *prefix)
+{
+  char buf[64];
+  const char *addr = buf;
+
+  if (net_address_get(buf, sizeof(buf), naddr) < 0)
+    return false;
+
+  if (naddr->sa.sa_family == AF_INET6 && IN6_IS_ADDR_V4MAPPED(&naddr->sin6.sin6_addr))
+    addr += strlen("::ffff:");
+
+  if (strncmp(addr, prefix, strlen(prefix)) == 0)
+    return true;
+
+  return false;
+}
+
 bool
-net_peer_address_is_trusted(const char *addr)
+net_peer_address_is_trusted(union net_sockaddr *naddr)
 {
   cfg_t *section;
   const char *network;
   int i;
   int n;
 
-  if (!addr)
+  if (!naddr)
     return false;
 
-  if (strncmp(addr, "::ffff:", strlen("::ffff:")) == 0)
-    addr += strlen("::ffff:");
-
   section = cfg_getsec(cfg, "general");
 
   n = cfg_size(section, "trusted_networks");
@@ -141,17 +240,20 @@ net_peer_address_is_trusted(const char *addr)
     {
       network = cfg_getnstr(section, "trusted_networks", i);
 
-      if (!network || network[0] == '\0')
+      if (!network || network[0] == '\0' || strcmp(network, "none") == 0)
 	return false;
 
-      if (strncmp(network, addr, strlen(network)) == 0)
-	return true;
-
-      if ((strcmp(network, "localhost") == 0) && (strcmp(addr, "127.0.0.1") == 0 || strcmp(addr, "::1") == 0))
-	return true;
-
       if (strcmp(network, "any") == 0)
 	return true;
+
+      if (strcmp(network, "lan") == 0 && net_address_is_local(naddr))
+	return true;
+
+      if (strcmp(network, "localhost") == 0)
+	network = (naddr->sa.sa_family == AF_INET6) ? "::1" : "127.0.0.1";
+
+      if (net_address_has_prefix(naddr, network))
+	return true;
     }
 
   return false;
@@ -160,13 +262,13 @@ net_peer_address_is_trusted(const char *addr)
 int
 net_address_get(char *addr, size_t addr_len, union net_sockaddr *naddr)
 {
-  const char *s;
+  const char *s = NULL;
 
   memset(addr, 0, addr_len); // Just in case caller doesn't check for errors
 
   if (naddr->sa.sa_family == AF_INET6)
      s = inet_ntop(AF_INET6, &naddr->sin6.sin6_addr, addr, addr_len);
-  else
+  else if (naddr->sa.sa_family == AF_INET)
      s = inet_ntop(AF_INET, &naddr->sin.sin_addr, addr, addr_len);
 
   if (!s)
@@ -180,70 +282,186 @@ net_port_get(short unsigned *port, union net_sockaddr *naddr)
 {
   if (naddr->sa.sa_family == AF_INET6)
      *port = ntohs(naddr->sin6.sin6_port);
-  else
+  else if (naddr->sa.sa_family == AF_INET)
      *port = ntohs(naddr->sin.sin_port);
+  else
+    return -1;
 
   return 0;
 }
 
 int
-net_connect(const char *addr, unsigned short port, int type, const char *log_service_name)
+net_if_get(char *ifname, size_t ifname_len, const char *addr)
+{
+  struct ifaddrs *ifaddrs;
+  struct ifaddrs *iap;
+  char s[64];
+
+  memset(ifname, 0, ifname_len);
+
+  getifaddrs(&ifaddrs);
+
+  for (iap = ifaddrs; iap; iap = iap->ifa_next)
+    {
+      if (!iap->ifa_addr)
+	continue;
+
+      if (net_address_get(s, sizeof(s), (union net_sockaddr *)iap->ifa_addr) < 0)
+	continue;
+
+      if (strcmp(s, addr) != 0)
+	continue;
+
+      snprintf(ifname, ifname_len, "%s", iap->ifa_name);
+      break;
+    }
+
+  freeifaddrs(ifaddrs);
+
+  return (ifname[0] != 0) ? 0 : -1;
+}
+
+static int
+net_connect_addrinfo(struct addrinfo *ai, int timeout_ms, bool set_nonblock, const char **errmsg)
+{
+  int fd = -1;
+  int flags;
+  struct pollfd pollfd;
+  socklen_t len;
+  int error;
+  int ret;
+
+  fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
+  if (fd < 0)
+    {
+      *errmsg = strerror(errno);
+      goto error;
+    }
+
+  // For Linux we could just give SOCK_CLOEXEC to socket(), but that won't work
+  // with MacOS, so we have to use fcntl()
+  flags = fcntl(fd, F_GETFL, 0);
+  if (flags < 0)
+    {
+      *errmsg = "fcntl() with F_GETFL returned an error";
+      goto error;
+    }
+
+  ret = fcntl(fd, F_SETFL, flags | O_NONBLOCK | O_CLOEXEC);
+  if (ret < 0)
+    {
+      *errmsg = "fcntl() with F_SETFL non-block returned an error";
+      goto error;
+    }
+
+  // We often need to wait for the connection. On Linux this seems always to be
+  // the case, but FreeBSD connect() sometimes returns immediate success.
+  ret = connect(fd, ai->ai_addr, ai->ai_addrlen);
+  if (ret < 0 && errno != EINPROGRESS) // EINPROGRESS in case of nonblock
+    {
+      *errmsg = strerror(errno);
+      goto error;
+    }
+  else if (ret != 0)
+    {
+      // Use poll here since select requires using fdset that would be
+      // overflowed in FreeBSD
+      pollfd.fd = fd;
+      pollfd.events = POLLOUT;
+
+      ret = poll(&pollfd, 1, timeout_ms);
+      if (ret < 0)
+	{
+	  *errmsg = strerror(errno);
+	  goto error;
+	}
+      else if (ret == 0)
+	{
+	  *errmsg = "Timed out trying to connect";
+	  goto error;
+	}
+
+      len = sizeof(error);
+      ret = getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len);
+      if (ret < 0 || error)
+	{
+	  *errmsg = error ? strerror(error) : strerror(errno);
+	  goto error;
+	}
+    }
+
+  if (!set_nonblock)
+    {
+      ret = fcntl(fd, F_SETFL, flags | O_CLOEXEC);
+      if (ret < 0)
+	{
+	  *errmsg = "fcntl() with F_SETFL block returned an error";
+	  goto error;
+	}
+    }
+
+  return fd;
+
+ error:
+  if (fd >= 0)
+    close(fd);
+
+  return -1;
+}
+
+static int
+net_connect_impl(const char *addr, unsigned short port, int type, const char *log_service_name, int timeout_ms, bool set_nonblock)
 {
   struct addrinfo hints = { 0 };
   struct addrinfo *servinfo;
-  struct addrinfo *ptr;
+  struct addrinfo *ai;
   char strport[8];
-  int fd;
+  const char *errmsg = NULL;
+  int fd = -1;
   int ret;
 
   DPRINTF(E_DBG, L_MISC, "Connecting to '%s' at %s (port %u)\n", log_service_name, addr, port);
 
-  hints.ai_socktype = (type & (SOCK_STREAM | SOCK_DGRAM)); // filter since type can be SOCK_STREAM | SOCK_NONBLOCK
+  hints.ai_socktype = type;
   hints.ai_family = (cfg_getbool(cfg_getsec(cfg, "general"), "ipv6")) ? AF_UNSPEC : AF_INET;
 
   snprintf(strport, sizeof(strport), "%hu", port);
   ret = getaddrinfo(addr, strport, &hints, &servinfo);
   if (ret < 0)
     {
-      DPRINTF(E_LOG, L_MISC, "Could not connect to '%s' at %s (port %u): %s\n", log_service_name, addr, port, gai_strerror(ret));
+      DPRINTF(E_LOG, L_MISC, "Could not get '%s' address info for %s (port %u): %s\n", log_service_name, addr, port, gai_strerror(ret));
       return -1;
     }
 
-  for (ptr = servinfo; ptr; ptr = ptr->ai_next)
+  for (ai = servinfo; ai && fd < 0; ai = ai->ai_next)
     {
-      fd = socket(ptr->ai_family, type | SOCK_CLOEXEC, ptr->ai_protocol);
-      if (fd < 0)
-	{
-	  continue;
-	}
-
-      ret = connect(fd, ptr->ai_addr, ptr->ai_addrlen);
-      if (ret < 0 && errno != EINPROGRESS) // EINPROGRESS in case of SOCK_NONBLOCK
-	{
-	  close(fd);
-	  continue;
-	}
-
-      break;
+      fd = net_connect_addrinfo(ai, timeout_ms, set_nonblock, &errmsg);
     }
 
   freeaddrinfo(servinfo);
 
-  if (!ptr)
+  if (fd < 0)
     {
-      DPRINTF(E_LOG, L_MISC, "Could not connect to '%s' at %s (port %u): %s\n", log_service_name, addr, port, strerror(errno));
+      DPRINTF(E_LOG, L_MISC, "Could not connect to '%s' at %s (port %u): %s\n", log_service_name, addr, port, errmsg);
       return -1;
     }
 
-  // net_address_get(ipaddr, sizeof(ipaddr), (union net_sockaddr *)ptr->ai-addr);
+  // net_address_get(ipaddr, sizeof(ipaddr), (union net_sockaddr *)ai->ai-addr);
 
   return fd;
 }
 
-// If *port is 0 then a random port will be assigned, and *port will be updated
-// with the port number
 int
-net_bind(short unsigned *port, int type, const char *log_service_name)
+net_connect(const char *addr, unsigned short port, int type, const char *log_service_name)
+{
+  return net_connect_impl(addr, port, type, log_service_name, NET_CONNECT_TIMEOUT_MS, false);
+}
+
+// If *port is 0 then a random port will be assigned, and *port will be updated
+// with the port number. SOCK_STREAM type services are set to use non-blocking
+// sockets.
+static int
+net_bind_impl(short unsigned *port, int type, const char *log_service_name, bool reuseport)
 {
   struct addrinfo hints = { 0 };
   struct addrinfo *servinfo;
@@ -253,6 +471,7 @@ net_bind(short unsigned *port, int type, const char *log_service_name)
   const char *cfgaddr;
   char addr[INET6_ADDRSTRLEN];
   char strport[8];
+  int flags;
   int yes = 1;
   int no = 0;
   int fd = -1;
@@ -260,7 +479,7 @@ net_bind(short unsigned *port, int type, const char *log_service_name)
 
   cfgaddr = cfg_getstr(cfg_getsec(cfg, "general"), "bind_address");
 
-  hints.ai_socktype = (type & (SOCK_STREAM | SOCK_DGRAM)); // filter since type can be SOCK_STREAM | SOCK_NONBLOCK
+  hints.ai_socktype = type;
   hints.ai_family = (cfg_getbool(cfg_getsec(cfg, "general"), "ipv6")) ? AF_INET6 : AF_INET;
   hints.ai_flags = cfgaddr ? 0 : AI_PASSIVE;
 
@@ -277,10 +496,30 @@ net_bind(short unsigned *port, int type, const char *log_service_name)
       if (fd >= 0)
 	close(fd);
 
-      fd = socket(ptr->ai_family, type | SOCK_CLOEXEC, ptr->ai_protocol);
+      fd = socket(ptr->ai_family, ptr->ai_socktype, ptr->ai_protocol);
       if (fd < 0)
 	continue;
 
+      // For Linux we could just give SOCK_NONBLOCK and SOCK_CLOEXEC to
+      // socket(), but that won't work with MacOS, so we have to use fcntl()
+      flags = fcntl(fd, F_GETFL, 0);
+      if (flags < 0)
+	continue;
+      if (type == SOCK_STREAM)
+	ret = fcntl(fd, F_SETFL, flags | O_NONBLOCK | O_CLOEXEC);
+      else
+	ret = fcntl(fd, F_SETFL, flags | O_CLOEXEC);
+      if (ret < 0)
+	continue;
+
+      // Makes us able to attach multiple threads to the same port
+      if (reuseport)
+	ret = setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &yes, sizeof(yes));
+      else
+	ret = setsockopt(fd, SOL_SOCKET, SO_REUSEPORT, &no, sizeof(no));
+      if (ret < 0)
+	continue;
+
       // TODO libevent sets this, we do the same?
       ret = setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &yes, sizeof(yes));
       if (ret < 0)
@@ -341,7 +580,19 @@ net_bind(short unsigned *port, int type, const char *log_service_name)
 }
 
 int
-net_evhttp_bind(struct evhttp *evhttp, short unsigned port, const char *log_service_name)
+net_bind(short unsigned *port, int type, const char *log_service_name)
+{
+  return net_bind_impl(port, type, log_service_name, false);
+}
+
+int
+net_bind_with_reuseport(short unsigned *port, int type, const char *log_service_name)
+{
+  return net_bind_impl(port, type, log_service_name, true);
+}
+
+int
+net_evhttp_bind(struct evhttp *evhttp, unsigned short port, const char *log_service_name)
 {
   const char *bind_address;
   bool v6_enabled;
@@ -408,19 +659,19 @@ safe_atoi32(const char *str, int32_t *val)
   if (((errno == ERANGE) && ((intval == LONG_MAX) || (intval == LONG_MIN)))
       || ((errno != 0) && (intval == 0)))
     {
-      DPRINTF(E_DBG, L_MISC, "Invalid integer in string (%s): %s\n", str, strerror(errno));
+      DPRINTF(E_DBG, L_MISC, "Invalid i32 in string (%s): %s\n", str, strerror(errno));
       return -1;
     }
 
   if (end == str)
     {
-      DPRINTF(E_DBG, L_MISC, "No integer found in string (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "No i32 found in string (%s)\n", str);
       return -1;
     }
 
   if (intval > INT32_MAX)
     {
-      DPRINTF(E_DBG, L_MISC, "Integer value too large (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "i32 value out of range (%s)\n", str);
       return -1;
     }
 
@@ -447,19 +698,19 @@ safe_atou32(const char *str, uint32_t *val)
   if (((errno == ERANGE) && (intval == ULONG_MAX))
       || ((errno != 0) && (intval == 0)))
     {
-      DPRINTF(E_DBG, L_MISC, "Invalid integer in string (%s): %s\n", str, strerror(errno));
+      DPRINTF(E_DBG, L_MISC, "Invalid u32 in string (%s): %s\n", str, strerror(errno));
       return -1;
     }
 
   if (end == str)
     {
-      DPRINTF(E_DBG, L_MISC, "No integer found in string (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "No u32 found in string (%s)\n", str);
       return -1;
     }
 
   if (intval > UINT32_MAX)
     {
-      DPRINTF(E_DBG, L_MISC, "Integer value too large (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "u32 value out of range (%s)\n", str);
       return -1;
     }
 
@@ -486,19 +737,19 @@ safe_hextou32(const char *str, uint32_t *val)
   if (((errno == ERANGE) && (intval == ULONG_MAX))
       || ((errno != 0) && (intval == 0)))
     {
-      DPRINTF(E_DBG, L_MISC, "Invalid integer in string (%s): %s\n", str, strerror(errno));
+      DPRINTF(E_DBG, L_MISC, "Invalid u32 in string (%s): %s\n", str, strerror(errno));
       return -1;
     }
 
   if (end == str)
     {
-      DPRINTF(E_DBG, L_MISC, "No integer found in string (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "No u32 found in string (%s)\n", str);
       return -1;
     }
 
   if (intval > UINT32_MAX)
     {
-      DPRINTF(E_DBG, L_MISC, "Integer value too large (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "u32 value out of range (%s)\n", str);
       return -1;
     }
 
@@ -525,19 +776,19 @@ safe_atoi64(const char *str, int64_t *val)
   if (((errno == ERANGE) && ((intval == LLONG_MAX) || (intval == LLONG_MIN)))
       || ((errno != 0) && (intval == 0)))
     {
-      DPRINTF(E_DBG, L_MISC, "Invalid integer in string (%s): %s\n", str, strerror(errno));
+      DPRINTF(E_DBG, L_MISC, "Invalid i64 in string (%s): %s\n", str, strerror(errno));
       return -1;
     }
 
   if (end == str)
     {
-      DPRINTF(E_DBG, L_MISC, "No integer found in string (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "No i64 found in string (%s)\n", str);
       return -1;
     }
 
   if (intval > INT64_MAX)
     {
-      DPRINTF(E_DBG, L_MISC, "Integer value too large (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "i64 value out of range (%s)\n", str);
       return -1;
     }
 
@@ -564,21 +815,21 @@ safe_atou64(const char *str, uint64_t *val)
   if (((errno == ERANGE) && (intval == ULLONG_MAX))
       || ((errno != 0) && (intval == 0)))
     {
-      DPRINTF(E_DBG, L_MISC, "Invalid integer in string (%s): %s\n", str, strerror(errno));
+      DPRINTF(E_DBG, L_MISC, "Invalid u64 in string (%s): %s\n", str, strerror(errno));
 
       return -1;
     }
 
   if (end == str)
     {
-      DPRINTF(E_DBG, L_MISC, "No integer found in string (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "No u64 found in string (%s)\n", str);
 
       return -1;
     }
 
   if (intval > UINT64_MAX)
     {
-      DPRINTF(E_DBG, L_MISC, "Integer value too large (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "u64 value out of range (%s)\n", str);
       return -1;
     }
 
@@ -605,19 +856,19 @@ safe_hextou64(const char *str, uint64_t *val)
   if (((errno == ERANGE) && (intval == ULLONG_MAX))
       || ((errno != 0) && (intval == 0)))
     {
-      DPRINTF(E_DBG, L_MISC, "Invalid integer in string (%s): %s\n", str, strerror(errno));
+      DPRINTF(E_DBG, L_MISC, "Invalid u64 in string (%s): %s\n", str, strerror(errno));
       return -1;
     }
 
   if (end == str)
     {
-      DPRINTF(E_DBG, L_MISC, "No integer found in string (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "No u64 found in string (%s)\n", str);
       return -1;
     }
 
   if (intval > UINT64_MAX)
     {
-      DPRINTF(E_DBG, L_MISC, "Integer value too large (%s)\n", str);
+      DPRINTF(E_DBG, L_MISC, "u64 value out of range (%s)\n", str);
       return -1;
     }
 
@@ -1544,6 +1795,40 @@ quality_is_equal(struct media_quality *a, struct media_quality *b)
   return (a->sample_rate == b->sample_rate && a->bits_per_sample == b->bits_per_sample && a->channels == b->channels && a->bit_rate == b->bit_rate);
 }
 
+enum media_format
+media_format_from_string(const char *s)
+{
+  if (strcmp(s, "pcm") == 0)
+    return MEDIA_FORMAT_PCM;
+  if (strcmp(s, "wav") == 0)
+    return MEDIA_FORMAT_WAV;
+  if (strcmp(s, "mp3") == 0)
+    return MEDIA_FORMAT_MP3;
+  if (strcmp(s, "alac") == 0)
+    return MEDIA_FORMAT_ALAC;
+  if (strcmp(s, "opus") == 0)
+    return MEDIA_FORMAT_OPUS;
+
+  return MEDIA_FORMAT_UNKNOWN;
+}
+
+const char *
+media_format_to_string(enum media_format format)
+{
+  if (format == MEDIA_FORMAT_PCM)
+    return "pcm";
+  if (format == MEDIA_FORMAT_WAV)
+    return "wav";
+  if (format == MEDIA_FORMAT_MP3)
+    return "mp3";
+  if (format == MEDIA_FORMAT_ALAC)
+    return "alac";
+  if (format == MEDIA_FORMAT_OPUS)
+    return "opus";
+
+  return "unknown";
+}
+
 
 /* -------------------------- Misc utility functions ------------------------ */
 
@@ -1567,6 +1852,42 @@ mutex_init(pthread_mutex_t *mutex)
   return err;
 }
 
+int
+thread_gettid()
+{
+  int tid = -1;
+#if defined(HAVE_GETTID)
+  tid = (int)gettid();
+#elif defined(HAVE_PTHREAD_GETTHREADID_NP)
+  tid = pthread_getthreadid_np();
+#endif
+  return tid;
+}
+
+void
+thread_getname(pthread_t thread, char *name, size_t len)
+{
+#if defined(HAVE_PTHREAD_GETNAME_NP)
+  pthread_getname_np(thread, name, len);
+#elif defined(HAVE_PTHREAD_GET_NAME_NP)
+  pthread_get_name_np(thread, name, len);
+#else
+  name[0] = '\0';
+#endif
+}
+
+void
+thread_getnametid(char *buf, size_t len)
+{
+  int tid;
+  char thread_name[32];
+  pthread_t p = pthread_self();
+
+  thread_getname(p, thread_name, sizeof(thread_name));
+  tid = thread_gettid() % 10000;
+  snprintf(buf, len, "%s (%d)", thread_name, tid);
+}
+
 void
 thread_setname(pthread_t thread, const char *name)
 {
