@@ -861,7 +861,7 @@ rcp_session_shutdown_init(struct rcp_session* s)
     }
   else
     {
-      DPRINTF(E_DBG, L_RCP, "Limiting shutdown timeout %ld sec '%s' at %s\n", clear_timeout.tv_sec, s->devname, s->address);
+      DPRINTF(E_DBG, L_RCP, "Limiting shutdown timeout %ld sec '%s' at %s\n", (long)clear_timeout.tv_sec, s->devname, s->address);
 
       // ensure we're not blocked forever on responses
       event_add(s->reply_timeout, &clear_timeout);
@@ -1295,6 +1295,12 @@ rcp_mdns_device_cb(const char *name, const char *type, const char *domain, const
   device->name = strdup(name);
   device->type = OUTPUT_TYPE_RCP;
   device->type_name = outputs_name(device->type);
+  device->default_format = MEDIA_FORMAT_WAV;
+  device->supported_formats = MEDIA_FORMAT_WAV | MEDIA_FORMAT_MP3 | MEDIA_FORMAT_ALAC;
+
+  // RCP/Roku Soundbridges only support ipv4
+  device->v4_address = safe_strdup(address);
+  device->v4_port = port; // -1 if we are removing
 
   if (port < 0 || !address)
     {
@@ -1302,12 +1308,7 @@ rcp_mdns_device_cb(const char *name, const char *type, const char *domain, const
     }
   else
     {
-      // RCP/Roku Soundbridges only support ipv4
-      device->v4_address = strdup(address);
-      device->v4_port = port;
-
       DPRINTF(E_INFO, L_RCP, "Adding RCP output device '%s' at '%s'\n", name, address);
-
       ret = player_device_add(device);
     }
 
